"""
Base class for strategy packs.
"""
from ..utils import get_func_name


class StrategyPack(object):
    def __init__(self, initial_strats, inferral_strats, expansion_strats,
                 ver_strats, name, **kwargs):
        if initial_strats is None:
            raise TypeError(("Strategy pack requires a (possibly empty) list "
                             "of equivalence strategies."))
        elif ver_strats is None:
            raise TypeError(("Strategy pack requires a (possibly empty) list "
                             "of verification strategies."))
        elif inferral_strats is None:
            raise TypeError(("Strategy pack requires a (possibly empty) list "
                             "of inferral strategies."))
        elif expansion_strats is None:
            raise TypeError(("Strategy pack requires a (possibly empty) list "
                             "of lists of other strategies."))
        self.name = name
        self.initial_strats = initial_strats
        self.inferral_strats = inferral_strats
        self.ver_strats = ver_strats
        self.expansion_strats = expansion_strats
        self.symmetries = kwargs.get('symmetries', [])
        self.forward_equivalence = kwargs.get('forward_equivalence', False)
        self.iterative = kwargs.get('iterative', False)

    def __str__(self):
        string = ("Looking for {} combinatorial specification"
                  " with the strategies:\n").format(
                        'iterative' if self.iterative else 'recursive')
        initial_strats = ", ".join(get_func_name(f)
                                   for f in self.initial_strats)
        infer_strats = ", ".join(get_func_name(f)
                                 for f in self.inferral_strats)
        verif_strats = ", ".join(get_func_name(f)
                                 for f in self.ver_strats)
        string += "Inferral: {}\n".format(infer_strats)
        string += "Initial: {}\n".format(initial_strats)
        string += "Verification: {}\n".format(verif_strats)
        if self.forward_equivalence:
            string += "Using forward equivalence only.\n"
        if self.symmetries:
            symme_strats = ", ".join(get_func_name(f)
                                     for f in self.symmetries)
            string += "Symmetries: {}\n".format(symme_strats)
        for i, strategies in enumerate(self.expansion_strats):
            strats = ", ".join(get_func_name(f) for f in strategies)
            string += "Set {}: {}\n".format(str(i+1), strats)
        return string



"""
A database to keep track of equivalent combinatorial classes.

This is done using a union find method. Also, explanations of how combinatorial
classes  are equivalent are maintained.

Based on: https://www.ics.uci.edu/~eppstein/PADS/UnionFind.py
"""

from collections import deque


class EquivalenceDB(object):
    """
    A database for equivalences. Supports four methods.

    - DB[x] return a name for the set containing x. Each set named by
    arbitrarily chosen member.
    - DB.union(t1, t2, explanation) merges the sets containing t1 and t2 and
    records why t1 and t2 are equivalent.
    - DB.equivalent(t1, t2) returns True, if t1 and t2 are in the same set,
    otherwise returns False.
    - DB.get_relation(t1, t2) returns a string explaining why t1 and t2 are
      equivalent.
    """

    def __init__(self):
        """Create a new empty equivalent database."""
        self.parents = {}
        self.weights = {}
        self.explanations = {}
        self.verified_roots = set()

    def __eq__(self, other):
        """Check if all information stored is the same."""
        return (self.parents == other.parents and
                self.weights == other.weights and
                self.explanations == other.explanations and
                self.verified_roots == other.verified_roots)

    def to_dict(self):
        """Return dictionary object of self that is JSON serializable."""
        return {
            'parents': self.parents,
            'weights': self.weights,
            'explanations': [(list(x), y)
                             for x, y in self.explanations.items()],
            'verified_roots': list(self.verified_roots),
        }

    @classmethod
    def from_dict(cls, dict):
        """Return EquivalenceDB object for dictionary object."""
        equivdb = cls()
        equivdb.parents = {int(x): y for x, y in dict['parents'].items()}
        equivdb.weights = {int(x): y for x, y in dict['weights'].items()}
        equivdb.explanations = {tuple(x): y
                                for x, y in dict['explanations']}
        equivdb.verified_roots = set(dict['verified_roots'])
        return equivdb

    def __getitem__(self, comb_class):
        """Find and return root combinatorial class for the set containing
        comb_class."""
        root = self.parents.get(comb_class)
        if root is None:
            self.parents[comb_class] = comb_class
            self.weights[comb_class] = 1
            return comb_class

        # Find path of combinatorial class leading to root.
        path = [comb_class]
        while root != path[-1]:
            path.append(root)
            root = self.parents[root]

        # Compress the path and return
        for ancestor in path:
            self.parents[ancestor] = root
        return root

    def union(self, t1, t2, explanation):
        """Find sets containing t1 and t2 and merge them."""
        verified = self.is_verified(t1) or self.is_verified(t2)
        roots = [self[t1], self[t2]]
        heaviest = max([(self.weights[r], r) for r in roots])[1]
        for r in roots:
            if r != heaviest:
                self.weights[heaviest] += self.weights[r]
                self.parents[r] = heaviest
        self.explanations[(t1, t2)] = explanation
        if (t2, t1) not in self.explanations:
            self.explanations[(t2, t1)] = "Reverse of: " + explanation
        if verified:
            self.update_verified(t1)

    def equivalent(self, t1, t2):
        """Return True if t1 and t2 are equivalent, False otherwise."""
        return self[t1] == self[t2]

    def update_verified(self, comb_class):
        """Update database that combinatorial classes equivalent to comb_class
        are verified."""
        if not self.is_verified(comb_class):
            self.verified_roots.add(self[comb_class])

    def is_verified(self, comb_class):
        """Return true if any equivalent combinatorial class is verified."""
        return self[comb_class] in self.verified_roots

    def equivalent_set(self, comb_class):
        equivalent_classes = set()
        for t in self.parents:
            if self.equivalent(comb_class, t):
                equivalent_classes.add(t)
        return equivalent_classes

    def get_explanation(self, comb_class, other_comb_class, one_step=False):
        """Return how two combinatorial classes are equivalent using
        explanations."""
        if comb_class == other_comb_class:
            return ""

        if one_step:
            explanation = self.explanations.get((comb_class, other_comb_class))
            if explanation is None:
                explanation = self.explanations.get((other_comb_class,
                                                     comb_class))
                assert explanation is not None
                explanation = "The reverse of: " + explanation
            return explanation

        path = self.find_path(comb_class, other_comb_class)
        if path:
            explanation = "| "
            for i in range(len(path) - 1):
                for j in range(i+1, len(path)):
                    t1 = path[i]
                    t2 = path[j]
                    key = (t1, t2)
                    if key in self.explanations:
                        new_explanation = self.explanations[key]
                        explanation = explanation + new_explanation + ". | "
                    key = (self[t2], self[t1])
                    if key in self.explanations:
                        new_explanation = self.explanations[key]
                        new_explanation = "The reverse of: " + new_explanation
                        explanation = explanation + new_explanation + ". | "
            return explanation
        raise KeyError("They are not equivalent.")

    def get_equiv_info(self, comb_class, other_comb_class):
        path = self.find_path(comb_class, other_comb_class)
        formal_step = self.get_explanation(comb_class, other_comb_class, path)

    def find_path(self, comb_class, other_comb_class):
        """
        BFS for shortest path.

        Used to find shortest explanation of why things are equivalent.
        """
        if not self.equivalent(comb_class, other_comb_class):
            raise KeyError("The classes given are not equivalent.")
        if comb_class == other_comb_class:
            return [comb_class]
        equivalent_comb_classes = {}
        reverse_map = {}

        for x in self.parents:
            n = len(equivalent_comb_classes)
            if self.equivalent(comb_class, x):
                equivalent_comb_classes[x] = n
                reverse_map[n] = x

        adjacency_list = [[] for i in range(len(equivalent_comb_classes))]
        for (t1, t2) in self.explanations:
            if self.equivalent(t1, comb_class):
                u = equivalent_comb_classes[t1]
                v = equivalent_comb_classes[t2]
                adjacency_list[u].append(v)
                adjacency_list[v].append(u)

        dequeue = deque()

        start = equivalent_comb_classes[comb_class]
        end = equivalent_comb_classes[other_comb_class]

        n = len(equivalent_comb_classes)

        dequeue.append(start)
        visited = [False for i in range(n)]
        neighbour = [None for i in range(n)]
        while len(dequeue) > 0:
            u = dequeue.popleft()
            if u == end:
                break
            if visited[u]:
                continue
            visited[u] = True

            for v in adjacency_list[u]:
                if not visited[v]:
                    dequeue.append(v)
                    neighbour[v] = u

        path = [reverse_map[end]]
        while neighbour[end] is not None:
            t = reverse_map[neighbour[end]]
            path.append(t)
            end = neighbour[end]

        return path[::-1]



from .comb_spec_searcher import CombinatorialSpecificationSearcher
from .combinatorial_class import CombinatorialClass
from .proof_tree import ProofTree
from .strategies import (BatchRule, BatchStrategy, DecompositionRule,
                         DecompositionStrategy, EquivalenceRule,
                         EquivalenceStrategy, InferralRule, InferralStrategy,
                         Rule, Strategy, StrategyPack, VerificationRule,
                         VerificationStrategy)



import abc

import sympy.abc

__all__ = ("CombinatorialClass")


class CombinatorialClass(abc.ABC):
    """
    Base class for CombSpecSearcher combinatorial class

    This is a base combinatorial class that CombinatorialSpecificationSearcher
    works with. Combinatorial classes to be used with the searcher should
    inherit from this class.
    """

    @abc.abstractmethod
    def is_empty(self, *args, **kwargs):
        """Return True if there are no object of any lengths."""
        return False

    @abc.abstractmethod
    def to_jsonable(self):
        """Return JSONable data structure of the class (a dictionary)"""
        pass

    def get_genf(self, *args, **kwargs):
        """Return the generating function for the combinatorial class"""
        raise NotImplementedError(("If you want to use the 'get_genf' function"
                                   "for a proof tree then you must implement"
                                   " 'get_genf' for verified combinatorial "
                                   "classes."))

    def get_min_poly(self, *args, **kwargs):
        """Return the minimum polynomial for the combinatorial class
        in terms of 'F'"""
        raise NotImplementedError(("If you want to use the 'get_min_poly' "
                                   "function for a proof tree then you must "
                                   "implement 'get_genf' for verified "
                                   "combinatorial classes."))

    def objects_of_length(self, length):
        """Returns an iterable of combinatorial objects of a given length"""
        raise NotImplementedError("This function needs to be added to your "
                                  "combinatorial class in order to use the "
                                  "debug settings and for initial conditions"
                                  " for computing the generating function")

    def is_epsilon(self):
        """Returns True if the generating function equals 1"""
        raise NotImplementedError("If you want to use the "
                                  "'count_objects_of_length' function "
                                  "for a proof tree then you must implement "
                                  "'is_epsilon' for your combinatorial class.")

    def is_atom(self):
        """Returns True if the generating function equals x"""
        raise NotImplementedError("If you want to use the "
                                  "'count_objects_of_length' function "
                                  "for a proof tree then you must implement "
                                  "'is_epsilon', 'is_atom' and 'is_positive' "
                                  "for your combinatorial class.")

    def is_positive(self):
        """Returns True if the constant term of the generating function is 0"""
        raise NotImplementedError("If you want to use the "
                                  "'count_objects_of_length' function "
                                  "for a proof tree then you must implement "
                                  "'is_epsilon', 'is_atom' and 'is_positive' "
                                  "for your combinatorial class.")

    def from_dict(self):
        """Return combinatorial class from the jsonable object."""
        raise NotImplementedError("This function is need to reinstantiate a "
                                  "combinatorial class.")

    def from_parts(self, *args, **kwargs):
        """Return the object created from the parts given by a cartesian rule
        applied on the comb_class. This is required for random sampling. The
        formal step will be passed as the kwarg 'formal_step'."""
        raise NotImplementedError("This function is needed to perform random"
                                  " sampling.")

    def compress(self):
        """Return a compressed version of the class. If you are having memory
        issues then implement this function and the decompress function such
        that 'cls.decompress(self.compress()) == self'"""
        return self

    @classmethod
    def decompress(cls, compressed):
        """Return decompressed class from string by compress function. If you
        are having memory issues then implement this function and the compress
        function such that 'cls.decompress(self.compress()) == self'"""
        return compressed

    @abc.abstractmethod
    def __init__(self):
        return

    @abc.abstractmethod
    def __eq__(self, other):
        return False

    @abc.abstractmethod
    def __hash__(self):
        pass

    @abc.abstractmethod
    def __repr__(self):
        pass

    @abc.abstractmethod
    def __str__(self):
        pass

    def _get_class_name(self):
        return type(self).__name__



"""
A proof tree class.

This can be used to get the generating function for the class.
"""
import json
import random
import sys
import warnings
from functools import reduce
from operator import add, mul

import sympy
from logzero import logger

from comb_spec_searcher.utils import compositions

from .tree_searcher import Node as tree_searcher_node
from .utils import (check_equation, check_poly, compositions, get_solution,
                    maple_equations, taylor_expand)


class ProofTreeNode(object):
    def __init__(self, label, eqv_path_labels, eqv_path_comb_classes,
                 eqv_explanations=[], children=[], strategy_verified=False,
                 decomposition=False, disjoint_union=False, recursion=False,
                 formal_step=""):
        self.label = label
        self.eqv_path_labels = eqv_path_labels
        self.eqv_path_comb_classes = eqv_path_comb_classes
        self.eqv_explanations = eqv_explanations
        self.children = children
        self.strategy_verified = strategy_verified
        self.decomposition = decomposition
        self.disjoint_union = disjoint_union
        self.recursion = recursion
        self.formal_step = formal_step
        self.sympy_function = None
        self.terms = []
        self.recurse_node = None
        self.genf = None

    @property
    def logger_kwargs(self):
        return {"processname": "ProofTreeNode {}".format(self.label)}

    def to_jsonable(self):
        output = dict()
        output['label'] = self.label
        output['eqv_path_labels'] = [x for x in self.eqv_path_labels]
        output['eqv_path_comb_classes'] = [x.to_jsonable()
                                           for x in self.eqv_path_comb_classes]
        output['eqv_explanations'] = [x for x in self.eqv_explanations]
        output['children'] = [child.to_jsonable() for child in self.children]
        output['strategy_verified'] = self.strategy_verified
        output['decomposition'] = self.decomposition
        output['disjoint_union'] = self.disjoint_union
        output['recursion'] = self.recursion
        output['formal_step'] = self.formal_step
        return output

    @classmethod
    def from_dict(cls, combclass, jsondict):
        if 'eqv_path_objects' in jsondict:
            warnings.warn(("The 'eqv_path_objects' label is deprecated. You "
                           "should change this to 'eqv_path_comb_classes"
                           " in the future."),
                          DeprecationWarning, stacklevel=2)
            jsondict['eqv_path_comb_classes'] = jsondict['eqv_path_objects']
        return cls(label=jsondict['label'],
                   eqv_path_labels=jsondict['eqv_path_labels'],
                   eqv_path_comb_classes=[
                                combclass.from_dict(x)
                                for x in jsondict['eqv_path_comb_classes']],
                   eqv_explanations=jsondict['eqv_explanations'],
                   children=[ProofTreeNode.from_dict(combclass, child)
                             for child in jsondict['children']],
                   strategy_verified=jsondict['strategy_verified'],
                   decomposition=jsondict['decomposition'],
                   disjoint_union=jsondict['disjoint_union'],
                   recursion=jsondict['recursion'],
                   formal_step=jsondict['formal_step'])

    @classmethod
    def from_json(cls, combclass, jsonstr):
        jsondict = json.loads(jsonstr)
        return cls.from_dict(combclass, jsondict)

    def _error_string(self, parent, children, strat_type, formal_step,
                      length, parent_total, children_total):
        error = "Insane " + strat_type + " Strategy Found!\n"
        error += formal_step + "\n"
        error += "Found at length {} \n".format(length)
        error += "The parent comb class was:\n{}\n".format(parent.__repr__())
        error += "It produced {} many things\n".format(parent_total)
        error += "The children were:\n"
        for comb_class in children:
            error += comb_class.__repr__()
            error += "\n"
        error += "They produced {} many things\n\n".format(children_total)
        return error

    def random_sample(self, length, tree=None):
        """Return a random object of the given length."""
        def partitions(n, children_totals):
            if n == 0 and not children_totals:
                yield []
                return
            if len(children_totals) == 0 or n < 0:
                return
            start = children_totals[0]
            if len(children_totals) == 1:
                if start[n] != 0:
                    yield [n]
                return
            for i in range(n + 1):
                if start[i] == 0:
                    continue
                else:
                    for part in partitions(n - i, children_totals[1:]):
                        yield [i] + part

        if self.disjoint_union:
            total = self.terms[length]
            if total == 0:
                raise ValueError(("There is no object on node {} of length {}"
                                  "".format(self.label, length)))
            children_totals = [(child, child.terms[length])
                               for child in self.children]
            choice = random.randint(1, total)
            # TODO: consider if there is an equivalent path to follow
            # TODO: Add functionality for bijection implied by rule
            sofar = 0
            for child, child_total in children_totals:
                sofar += child_total
                if choice <= sofar:
                    return child.random_sample(length, tree)
            raise ValueError("You shouldn't be able to get here!")
        elif self.decomposition:
            total = self.terms[length]
            choice = random.randint(1, total)
            children_totals = [child.terms for child in self.children]
            sofar = 0
            for part in partitions(length, children_totals):
                subtotal = 1
                for i, terms in zip(part, children_totals):
                    subtotal *= terms[i]
                sofar += subtotal
                if choice <= sofar:
                    sub_objs = [(child.random_sample(i, tree),
                                 child.eqv_path_comb_classes[0])
                                for i, child in zip(part, self.children)]
                    comb_class = self.eqv_path_comb_classes[-1]
                    return comb_class.from_parts(*sub_objs,
                                                 formal_step=self.formal_step)
            raise ValueError("You shouldn't be able to get here.")
        elif self.strategy_verified:
            return self.eqv_path_comb_classes[-1].random_sample(length)
        else:
            if self.recursion:
                for node in tree.nodes():
                    if node.label == self.label and not node.recursion:
                        return node.random_sample(length, tree)
            raise NotImplementedError(("Random sampler only implemented for "
                                       "disjoint union and cartesian "
                                       "product."))

    def sanity_check(self, length, of_length=None):
        if of_length is None:
            raise ValueError("of_length is undefined.")

        number_objs = of_length(self.eqv_path_comb_classes[0], length)
        for i, comb_class in enumerate(self.eqv_path_comb_classes[1:]):
            eqv_number = of_length(comb_class, length)
            if number_objs != eqv_number:
                formal_step = ""
                for i in range(i+1):
                    formal_step += self.eqv_explanations[i]
                return self._error_string(self.eqv_path_comb_classes[0],
                                          [comb_class],
                                          "Equivalent",
                                          formal_step,
                                          length,
                                          number_objs,
                                          eqv_number)
        if self.disjoint_union:
            child_comb_classes = [child.eqv_path_comb_classes[0]
                                  for child in self.children]
            total = 0
            for comb_class in child_comb_classes:
                total += of_length(comb_class, length)
            if number_objs != total:
                return self._error_string(self.eqv_path_comb_classes[0],
                                          child_comb_classes,
                                          "Batch",
                                          self.formal_step,
                                          length,
                                          number_objs,
                                          total)
        if self.decomposition:
            child_comb_classes = [child.eqv_path_comb_classes[0]
                                  for child in self.children]
            total = 0
            for part in compositions(length, len(child_comb_classes)):
                subtotal = 1
                for comb_class, partlen in zip(child_comb_classes, part):
                    if subtotal == 0:
                        break
                    subtotal *= of_length(comb_class, partlen)
                total += subtotal
            if number_objs != total:
                return self._error_string(self.eqv_path_comb_classes[0],
                                          child_comb_classes,
                                          "Decomposition",
                                          self.formal_step,
                                          length,
                                          number_objs,
                                          total)

    def __eq__(self, other):
        return all([self.label == other.label,
                    self.eqv_path_labels == other.eqv_path_labels,
                    self.eqv_path_comb_classes == other.eqv_path_comb_classes,
                    self.eqv_explanations == other.eqv_explanations,
                    self.children == other.children,
                    self.strategy_verified == other.strategy_verified,
                    self.decomposition == other.decomposition,
                    self.disjoint_union == other.disjoint_union,
                    self.recursion == other.recursion,
                    self.formal_step == other.formal_step])

    def get_function(self, min_poly=False):
        if min_poly:
            if (self.sympy_function is None or
                    isinstance(self.sympy_function, sympy.Function)):
                self.sympy_function = sympy.var("F_" + str(self.label))
        else:
            if (self.sympy_function is None or
                    isinstance(self.sympy_function, sympy.Symbol)):
                self.sympy_function = sympy.Function(
                                        "F_" + str(self.label))(sympy.abc.x)
        return self.sympy_function

    def get_equation(self, root_func=None, root_class=None,
                     dummy_eq=False, min_poly=False, **kwargs):
        lhs = self.get_function(min_poly)
        if self.disjoint_union:
            rhs = reduce(add,
                         [child.get_function(min_poly)
                          for child in self.children],
                         0)
        elif self.decomposition:
            rhs = reduce(mul,
                         [child.get_function(min_poly)
                          for child in self.children],
                         1)
        elif self.recursion:
            rhs = lhs
        elif self.strategy_verified:
            comb_class = self.eqv_path_comb_classes[-1]
            try:
                if min_poly:
                    lhs = comb_class.get_min_poly(root_func=root_func,
                                                  root_class=root_class)
                    F = sympy.Symbol("F")
                    lhs = lhs.subs({F: self.get_function(min_poly)})
                    rhs = 0
                else:
                    rhs = comb_class.get_genf(root_func=root_func,
                                              root_class=root_class)
            except ValueError as e:
                if not dummy_eq:
                    raise ValueError(e)
                logger.info(("Unable to find equation for {}, adding dummy"
                             "function. The comb class corresponding is\n{}"
                             "".format(lhs, comb_class)),
                            extra=self.logger_kwargs)
                rhs = sympy.Function("DOITYOURSELF")(sympy.abc.x)
        else:
            if not dummy_eq:
                raise NotImplementedError("Using an unimplemented constructor")
            logger.info(("Unable to find equation for {}, adding dummy "
                         "function.".format(lhs)),
                        extra=self.logger_kwargs)
            rhs = sympy.Function("DOITYOURSELF")(sympy.abc.x)
        return sympy.Eq(lhs, rhs)

    def count_objects_of_length(self, n):
        '''
            Calculates objects of lenght in each node according to the
            recurrence relation implied by the proof tree. Only works
            for disjoint union, decomposition, strategy verified and recursion.

            Verified nodes are expected to have a known generating function.
        '''
        if n < 0:
            return 0
        if len(self.terms) > n and self.terms[n] is not None:
            return self.terms[n]

        ans = 0
        if self.disjoint_union:
            ans = sum(child.count_objects_of_length(n)
                      for child in self.children)
        elif self.decomposition:
            # Number of children that are just the atom
            atoms = 0
            # Indices of children that are positive (do not contain epsilon)
            pos_children = set()
            children = []  # A list of children that are not atoms
            for child in self.children:
                if child.eqv_path_comb_classes[-1].is_atom():
                    atoms += 1
                else:
                    if child.eqv_path_comb_classes[-1].is_positive():
                        pos_children.add(len(children))
                    children.append(child)

            for comp in compositions(n-atoms, len(children)):
                # A composition is only valid if all positive children
                # get more than 0 atoms.
                if any(c == 0 for i, c in enumerate(comp)
                        if i in pos_children):
                    continue
                tmp = 1
                for i, child in enumerate(children):
                    tmp *= child.count_objects_of_length(comp[i])
                    if tmp == 0:
                        break
                ans += tmp
        elif self.strategy_verified:
            if self.eqv_path_comb_classes[-1].is_epsilon():
                return 1 if n == 0 else 0
            elif self.eqv_path_comb_classes[-1].is_atom():
                return 1 if n == 1 else 0
            else:
                self._ensure_terms(n)
                return self.terms[n]
        elif self.recursion:
            if self.recurse_node:
                return self.recurse_node.count_objects_of_length(n)
            else:
                raise ValueError(("Recursing to a subtree that is not"
                                  " contained in the subtree from the"
                                  " root object that was called on."))
        else:
            raise NotImplementedError(("count_objects_of_length() is only "
                                       "defined for disjoint union, "
                                       "cartesian product, recursion "
                                       "and strategy verified."))
        if len(self.terms) <= n:
            self.terms.extend([None]*(n-len(self.terms)+1))
        self.terms[n] = ans
        return ans

    def _ensure_terms(self, n, expand_extra=50):
        """
            Ensures that self.terms contains the n-th term. If not it will
            use the generating function to compute terms up to n+expand_extra.
        """
        if len(self.terms) > n:
            return
        if self.genf is None:
            self.genf = self.eqv_path_comb_classes[-1].get_genf()
        coeffs = taylor_expand(self.genf, n=n+expand_extra)
        self.terms.extend(coeffs[len(self.terms):])

    @property
    def eqv_path_objects(self):
        """This is for reverse compatability."""
        warnings.warn(("The 'eqv_path_objects' label is deprecated. You "
                       "should change this to 'eqv_path_comb_classes"
                       " in the future."),
                      DeprecationWarning, stacklevel=2)
        return self.eqv_path_comb_classes


class InsaneTreeError(Exception):
    pass


class ProofTree(object):
    def __init__(self, root):
        if not isinstance(root, ProofTreeNode):
            raise TypeError("Root must be a ProofTreeNode.")
        self.root = root
        self._of_length_cache = {}
        self._fixed_recursion = False

    @property
    def logger_kwargs(self):
        return {'processname': 'ProofTree'}

    def to_jsonable(self):
        return {'root': self.root.to_jsonable()}

    @classmethod
    def from_dict(cls, combclass, jsondict):
        root = ProofTreeNode.from_dict(combclass, jsondict['root'])
        return cls(root)

    @classmethod
    def from_json(cls, combclass, jsonstr):
        jsondict = json.loads(jsonstr)
        return cls.from_dict(combclass, jsondict)

    def _of_length(self, comb_class, length):
        if comb_class not in self._of_length_cache:
            self._of_length_cache[comb_class] = {}

        number = self._of_length_cache[comb_class].get(length)

        if number is None:
            number = len(list(comb_class.objects_of_length(length)))
            self._of_length_cache[comb_class][length] = number

        return number

    def print_equivalences(self):
        """Return a string showing the equivalences."""
        s = ""
        for node in self.nodes():
            s += "===============\n"
            s += str(node.label) + "\n"
            for comb_class in node.eqv_path_comb_classes:
                s += str(comb_class) + "\n"

    def get_equations(self, dummy_eqs=False, min_poly=False, **kwargs):
        """Return the set of equations implied by the proof tree. If
        dummy_eqs=True then it will give a 'F_DOITYOURSELF' variable for
        equations that fail."""
        eqs = set()
        root_class = kwargs.get('root_class')
        root_func = kwargs.get('root_func')
        if root_class is None:
            root_class = self.root.eqv_path_comb_classes[0]
            kwargs['root_class'] = root_class
        if root_func is None:
            first_call = True
            root_func = sympy.Symbol("F_root")
            kwargs['root_func'] = root_func
        else:
            first_call = False
        for node in self.nodes():
            if node.recursion:
                continue
            eqs.add(node.get_equation(dummy_eq=dummy_eqs,
                                      min_poly=min_poly,
                                      **kwargs))
        if first_call:
            eqs.add(sympy.Eq(self.root.get_function(min_poly=True),
                             root_func))
        return eqs

    def get_genf(self, verify=8, only_root=True):
        """Find generating function for proof tree. Return None if no solution
        is found. If not verify will return list of possible solutions."""
        # TODO: add substitutions, so as to solve with symbols first.
        eqs = self.get_equations()
        root_class = self.root.eqv_path_comb_classes[0]
        root_func = self.root.get_function()
        eqs = self.get_equations(root_class=root_class, root_func=root_func)
        logger.info(maple_equations(root_func, root_class, eqs),
                    extra=self.logger_kwargs)
        logger.info("Solving...", extra=self.logger_kwargs)

        solutions = sympy.solve(eqs, tuple([eq.lhs for eq in eqs]),
                                dict=True, cubics=False, quartics=False,
                                quintics=False)

        if solutions:
            if not verify:
                return solutions
            logger.info("Solved, verifying solutions.",
                        extra=self.logger_kwargs)
            if only_root:
                objcounts = [len(list(root_class.objects_of_length(i)))
                             for i in range(verify + 1)]
                for solution in solutions:
                    genf = solution[root_func]
                    try:
                        expansion = taylor_expand(genf, verify)
                    except Exception as e:
                        continue
                    if objcounts == expansion:
                        logger.info(("The generating function is {}"
                                     "".format(genf)))
                        return genf
            else:
                for solution in solutions:
                    final_answer = {}
                    for node in self.nodes():
                        if node.label in final_answer:
                            continue
                        comb_class = node.eqv_path_comb_classes[0]
                        func = node.get_function()
                        objcounts = [len(list(comb_class.objects_of_length(i)))
                                     for i in range(verify + 1)]
                        genf = solution[func]
                        try:
                            expansion = taylor_expand(genf, verify)
                        except Exception as e:
                            continue
                        if objcounts == expansion:
                            logger.info(("The generating function for {} is {}"
                                         "".format(func, genf)))
                            final_answer[node.label] = genf
                        else:
                            break
                    else:
                        return final_answer
        if solutions:
            raise RuntimeError(("Incorrect generating function\n" +
                                str(solutions)))

    def get_min_poly(self, **kwargs):
        """Return the minimum polynomial of the generating function F that is
        implied by the proof tree."""
        root_class = kwargs.get('root_class')
        root_func = kwargs.get('root_func')
        if root_class is None:
            root_class = self.root.eqv_path_comb_classes[0]
            kwargs['root_class'] = root_class
        if root_func is None:
            first_call = True
            root_func = sympy.Symbol("F_root")
            kwargs['root_func'] = root_func
        else:
            first_call = False
        solve = kwargs.get('solve', False)

        eqs = self.get_equations(min_poly=True, **kwargs)

        func = self.root.get_function(min_poly=True)
        comb_class = self.root.eqv_path_comb_classes[0]
        logger.info(maple_equations(func, comb_class, eqs),
                    extra=self.logger_kwargs)
        logger.info("Computing Groebner basis with 'elimination' order.",
                    extra=self.logger_kwargs)
        all_funcs = set(x for eq in eqs for x in eq.atoms(sympy.Symbol))
        all_funcs.remove(func)
        if sympy.abc.x in all_funcs:
            all_funcs.remove(sympy.abc.x)

        elimination = sympy.polys.orderings.ProductOrder(
                (sympy.polys.orderings.grevlex, lambda m: m[:-2]),
                (sympy.polys.orderings.grevlex, lambda m: m[-2:])
        )

        basis = sympy.groebner(eqs, *all_funcs, sympy.abc.x, func,
                               order=elimination)
        eqs = basis.polys

        verify = 5
        if basis.polys:
            comb_class = self.root.eqv_path_comb_classes[0]
            initial = [len(list(comb_class.objects_of_length(i)))
                       for i in range(verify + 1)]
            if not first_call:
                root_initial = [len(list(root_class.objects_of_length(i)))
                                for i in range(verify + 1)]
                root_kwargs = {"root_func": root_func,
                               "root_initial": root_initial}

        F = sympy.Symbol("F")
        for poly in basis.polys:
            logger.debug(("Trying the minimum poly:\n{}\nwith the atoms\n{}\n"
                          "".format(poly.as_expr(), poly.atoms(sympy.Symbol))),
                         extra=self.logger_kwargs)
            if poly.atoms(sympy.Symbol) <= {func, sympy.abc.x}:
                logger.info("Trying the min poly:\n{}".format(poly.as_expr()),
                            extra=self.logger_kwargs)
                eq = poly.as_expr()
                eq = eq.subs({func: F})
                if check_poly(eq, initial) or check_equation(eq, initial):
                    logger.info(("The minimum polynomial is {}".format(eq)),
                                extra=self.logger_kwargs)
                    if solve:
                        sol = get_solution(eq, initial)
                        logger.info(("The generating function is {}"
                                     "".format(sol)), extra=self.logger_kwargs)
                        return eq, sol
                    else:
                        return eq
            elif poly.atoms(sympy.Symbol) <= {func, root_func, sympy.abc.x}:
                if first_call and sympy.abc.x not in poly.atoms(sympy.Symbol):
                    continue
                assert not solve
                eq = poly.as_expr()
                eq = eq.subs({func: F})
                if (check_poly(eq, initial, **root_kwargs) or
                        check_equation(eq, initial, **root_kwargs)):
                    logger.info(("The minimum polynomial is {}".format(eq)),
                                extra=self.logger_kwargs)
                    return eq
        raise RuntimeError(("Incorrect minimum polynomial\n" +
                            str(basis)))

    def random_sample(self, length=100, solved=False):
        if any(len(node.terms) < length + 1 for node in self.nodes()):
            logger.info(("Computing terms"))
            funcs = self.get_genf(only_root=False)
            for node in self.nodes():
                if len(node.terms) < length + 1:
                    logger.info(("Taylor expanding function {} to length {}."
                                 "".format(node.get_function(), length)))
                    node.terms = taylor_expand(funcs[node.label], length)
        logger.info("Walking through tree")
        return self.root.random_sample(length, self)

    def nodes(self, root=None):
        if root is None:
            root = self.root
        yield root
        for child in root.children:
            for node in self.nodes(root=child):
                yield node

    def number_of_nodes(self):
        return len(list(self.nodes()))

    def number_of_comb_classes(self):
        count = 0
        for node in self.nodes():
            count += len(node.eqv_path_comb_classes)
        return count

    def comb_classes(self, root=None):
        for node in self.nodes():
            for comb_class in node.eqv_path_comb_classes:
                yield comb_class

    def sanity_check(self, length=8, raiseerror=True):
        overall_error = ""
        for comb_class in self.comb_classes():
            if comb_class.is_empty():
                error = ("Combinatorial class is empty!\n{}\n"
                         "".format(repr(comb_class)))
                if raiseerror:
                    raise InsaneTreeError(error)
                else:
                    overall_error += error
        for node in self.nodes():
            error = node.sanity_check(length, self._of_length)
            if error is not None:
                if raiseerror:
                    raise InsaneTreeError(error)
                else:
                    overall_error += error
        if overall_error:
            return False, overall_error
        else:
            return True, "Sanity checked, all good at length {}".format(length)

    @classmethod
    def from_comb_spec_searcher(cls, root, css):
        if not isinstance(root, tree_searcher_node):
            raise TypeError("Requires a tree searcher node, treated as root.")
        proof_tree = ProofTree(ProofTree.from_comb_spec_searcher_node(root,
                                                                      css))
        proof_tree._recursion_fixer(css)
        return proof_tree

    def _recursion_fixer(self, css, root=None, in_labels=None):
        if root is None:
            root = self.root
        if in_labels is None:
            in_labels = list(self.non_recursive_in_labels())
        if root.recursion:
            in_label = root.eqv_path_labels[0]
            out_label = in_label
            for eqv_label in in_labels:
                if css.equivdb.equivalent(in_label, eqv_label):
                    out_label = eqv_label
                    break
            assert css.equivdb.equivalent(in_label, out_label)

            eqv_path = css.equivdb.find_path(in_label, out_label)
            eqv_comb_classes = [css.classdb.get_class(l) for l in eqv_path]
            eqv_explanations = [css.equivdb.get_explanation(x, y,
                                                            one_step=True)
                                for x, y in zip(eqv_path[:-1], eqv_path[1:])]

            root.eqv_path_labels = eqv_path
            root.eqv_path_comb_classes = eqv_comb_classes
            root.eqv_explanations = eqv_explanations

        for child in root.children:
            self._recursion_fixer(css, child, in_labels)

    def non_recursive_in_labels(self, root=None):
        if root is None:
            root = self.root
        if not root.recursion:
            yield root.eqv_path_labels[0]
        for child in root.children:
            for x in self.non_recursive_in_labels(child):
                yield x

    @classmethod
    def from_comb_spec_searcher_node(cls, root, css, in_label=None):
        if not isinstance(root, tree_searcher_node):
            raise TypeError("Requires a tree searcher node, treated as root.")
        label = root.label
        if in_label is None:
            in_label = root.label
        else:
            assert css.equivdb.equivalent(root.label, in_label)
        children = root.children

        if not children:
            eqv_ver_label = css.equivalent_strategy_verified_label(in_label)
            if eqv_ver_label is not None:
                # verified!
                eqv_path = css.equivdb.find_path(in_label, eqv_ver_label)
                eqv_comb_classes = [css.classdb.get_class(l) for l in eqv_path]
                eqv_explanations = [css.equivdb.get_explanation(x, y,
                                                                one_step=True)
                                    for x, y in zip(eqv_path[:-1],
                                                    eqv_path[1:])]

                formal_step = css.classdb.verification_reason(eqv_ver_label)
                return ProofTreeNode(label, eqv_path, eqv_comb_classes,
                                     eqv_explanations, strategy_verified=True,
                                     formal_step=formal_step)
            else:
                # recurse! we reparse these at the end, so recursed labels etc
                # are not interesting.
                return ProofTreeNode(label, [in_label],
                                     [css.classdb.get_class(in_label)],
                                     formal_step="recurse",
                                     recursion=True)
        else:
            rule = css.rule_from_equivence_rule(root.label,
                                                tuple(c.label
                                                      for c in root.children))
            start, ends = rule
            formal_step = css.ruledb.explanation(start, ends)
            constructor = css.ruledb.constructor(start, ends)

            eqv_path = css.equivdb.find_path(in_label, start)
            eqv_comb_classes = [css.classdb.get_class(l) for l in eqv_path]
            eqv_explanations = [css.equivdb.get_explanation(x, y,
                                                            one_step=True)
                                for x, y in zip(eqv_path[:-1], eqv_path[1:])]

            strat_children = []
            ends = list(ends)
            for child in root.children:
                for next_label in ends:
                    if css.equivdb.equivalent(next_label, child.label):
                        ends.remove(next_label)
                        sub_tree = ProofTree.from_comb_spec_searcher_node(
                                                        child, css, next_label)
                        strat_children.append(sub_tree)
                        break

            if constructor == 'cartesian':
                # decomposition!
                return ProofTreeNode(label, eqv_path, eqv_comb_classes,
                                     eqv_explanations, decomposition=True,
                                     formal_step=formal_step,
                                     children=strat_children)
            elif constructor == 'disjoint' or constructor == 'equiv':
                # batch!
                return ProofTreeNode(label, eqv_path, eqv_comb_classes,
                                     eqv_explanations, disjoint_union=True,
                                     formal_step=formal_step,
                                     children=strat_children)
            elif constructor == 'other':
                return ProofTreeNode(label, eqv_path, eqv_comb_classes,
                                     eqv_explanations,
                                     formal_step=formal_step,
                                     children=strat_children)
            else:
                logger.debug(("Unknown constructor '{}' of type '{}'. "
                              "Use 'other' instead."
                              "".format(constructor, type(constructor))),
                             extra={"processname": "css_to_proof_tree"})
                raise NotImplementedError("Only handle cartesian and disjoint")

    def _recursion_setup(self):
        label_to_node = dict()

        for node in self.nodes():
            if not node.recursion:
                label_to_node[node.label] = node

        for node in self.nodes():
            if node.recursion:
                node.recurse_node = label_to_node[node.label]

        self._fixed_recursion = True

    def count_objects_of_length(self, n):
        if not self._fixed_recursion:
            self._recursion_setup()
        return self.root.count_objects_of_length(n)

    def __eq__(self, other):
        return all(node1 == node2
                   for node1, node2 in zip(self.nodes(), other.nodes()))



"""
Finds and returns a combinatorial specification, that we call a proof tree.
"""
from collections import defaultdict, deque
from copy import deepcopy
from itertools import product
from random import choice, shuffle

__all__ = ("prune", "proof_tree_generator_dfs", "proof_tree_generator_bfs")


class Node(object):
    """A node for a proof tree."""
    def __init__(self, n, children=[]):
        self.label = n
        self.children = children

    def __str__(self):
        return "".join(["(", str(self.label), *map(str, self.children), ")"])

    def __len__(self):
        """Return the number nodes in the proof tree."""
        return 1 + sum(len(c) for c in self.children)


def prune(rules_dict):
    """Prune all nodes not in a combinatorial specification."""
    rdict = deepcopy(rules_dict)
    changed = True
    while changed:
        changed = False
        for k, rule_set in list(rdict.items()):
            for rule in list(rule_set):
                if any(x not in rdict for x in rule):
                    rule_set.remove(rule)
                    changed = True
                if not rule_set:
                    del rdict[k]
    return rdict


def iterative_prune(rules_dict, root=None):
    """Prune all nodes not iteratively verifiable."""
    verified_labels = set()
    if root is not None:
        verified_labels.add(root)
    rdict = deepcopy(rules_dict)
    new_rules_dict = defaultdict(set)
    while True:
        changed = False
        for k, rule_set in list(rdict.items()):
            for rule in list(rule_set):
                if all(x in verified_labels for x in rule):
                    changed = True
                    verified_labels.add(k)
                    new_rules_dict[k].add(rule)
                    rdict[k].remove(rule)
            if not rule_set:
                del rdict[k]
        if not changed:
            break
    return new_rules_dict


def proof_tree_dfs(rules_dict, root, seen=set()):
    """Return random proof tree found by depth first search."""
    seen = seen.copy()
    if root in rules_dict:
        rule_set = rules_dict[root]
        root_node = Node(root)
        if root in seen or () in rule_set:
            seen.add(root)
            return seen, root_node
        else:
            seen.add(root)
            rule = choice(list(rule_set))
            visited, trees = proof_forest_dfs(rules_dict, rule, seen)
            root_node.children = trees
            return visited, root_node


def proof_forest_dfs(rules_dict, roots, seen=set()):
    if not roots:
        return seen, []
    else:
        root, roots = roots[0], roots[1:]
        seen1, tree = proof_tree_dfs(rules_dict, root, seen)
        seen2, trees = proof_forest_dfs(rules_dict, roots, seen1)
        return seen1.union(seen2), [tree] + trees


# def proof_tree_generator_dfs(rules_dict, root):
#     """A generator for random proof trees using depth first search.
#     N.B. The rules_dict is assumed to be pruned.
#     """
#     while root in rules_dict:
#         _, tree = proof_tree_dfs(rules_dict, root)
#         yield tree


def iterative_proof_tree_bfs(rules_dict, root):
    """Takes in a iterative pruned rules_dict and returns iterative proof
    tree."""
    root_node = Node(root)
    queue = deque([root_node])
    while queue:
        v = queue.popleft()
        rule = sorted(rules_dict[v.label])[0]
        if not rule == ():
            children = [Node(i) for i in rule]
            queue.extend([child for child in children
                          if not child.label == root])
            v.children = children
    return root_node


def iterative_proof_tree_generator_bfs(rules_dict, root):
    pass


def random_proof_tree(rules_dict, root):
    """Return random tree found by breadth first search."""
    seen = set()
    root_node = Node(root)
    queue = deque([root_node])
    while queue:
        v = queue.popleft()
        rule = choice(list(rules_dict[v.label]))
        if not (v.label in seen or rule == ()):
            children = [Node(i) for i in rule]
            shuffle(children)
            queue.extend(children)
            v.children = children
        seen.add(v.label)
    return root_node


def proof_tree_generator_bfs(rules_dict, root):
    """A generator for all proof trees using breadth first search.
    N.B. The rules_dict is assumed to be pruned.
    """
    def _bfs_helper(root_label, seen):
        if root_label in seen:
            yield Node(root_label)
            return
        next_seen = seen.union((root_label,))
        for rule in rules_dict[root_label]:
            for children in product(*[_bfs_helper(child_label, next_seen)
                                      for child_label in rule]):
                root_node = Node(root_label)
                root_node.children = children
                yield root_node

    rules_dict = {start: tuple(sorted(ends))
                  for start, ends in rules_dict.items()}

    if root in rules_dict:
        yield from _bfs_helper(root, frozenset())


def proof_tree_generator_dfs(rules_dict, root, maximum=None):
    """A generator for all proof trees using depth first search.
    N.B. The rules_dict is assumed to be pruned.
    """
    def _dfs_tree(root_label, seen, maximum=None):
        if maximum is not None and maximum <= 0:
            return
        if root_label in seen:
            yield seen, Node(root_label)
            return
        seen = seen.union((root_label,))
        for rule in rules_dict[root_label]:
            if rule == ():
                yield seen, Node(root_label)
            else:
                for new_seen, children in _dfs_forest(rule, seen, maximum):
                    root_node = Node(root_label)
                    root_node.children = children
                    yield new_seen, root_node

    def _dfs_forest(root_labels, seen, maximum=None):
        if maximum is not None and maximum <= 0:
            return
        if not root_labels:
            yield seen, []
        else:
            root, roots = root_labels[0], root_labels[1:]
            for seen1, tree in _dfs_tree(root, seen,
                                         maximum - len(root_labels) + 1):
                length = len(tree)
                new_maximum = maximum - length if maximum is not None else None
                for seen2, trees in _dfs_forest(roots, seen1, new_maximum):
                    actual_length = length + sum(len(t) for t in trees)
                    if actual_length < maximum:
                        yield seen1.union(seen2), [tree] + trees

    rules_dict = {start: tuple(sorted(ends))
                  for start, ends in rules_dict.items()}

    if root in rules_dict:
        for _, tree in _dfs_tree(root, frozenset(), maximum):
            yield tree


def iterative_proof_tree_finder(rules_dict, root):
    """Finds an iterative proof tree for root, if one exists.
    """
    trees = {}

    def get_tree(start):
        if start == root:
            return Node(start)
        elif start in trees:
            return trees[start]
        else:
            raise KeyError("{} is not in trees".format(start))

    def create_tree(start, end):
        if start in trees:
            return
        root = Node(start)
        children = [get_tree(i) for i in end]
        root.children = children
        trees[start] = root

    verified_labels = set()
    if root is not None:
        verified_labels.add(root)
    rdict = deepcopy(rules_dict)
    new_rules_dict = defaultdict(set)
    while True:
        changed = False
        for k, rule_set in list(rdict.items()):
            for rule in list(rule_set):
                if all(x in verified_labels for x in rule):
                    changed = True
                    verified_labels.add(k)
                    new_rules_dict[k].add(rule)
                    create_tree(k, rule)
                    rdict[k].remove(rule)
            if not rule_set:
                del rdict[k]
        if not changed:
            break
    if root in trees:
        return trees[root]
    else:
        raise ValueError("{} has no tree in rules_dict".format(root))



"""
A database for combinatorial class found.

Contains information about if combinatorial classes have been expanded, found
by symmetries etc. It gives each combinatorial class a unique label.
"""

from base64 import b64decode, b64encode

from logzero import logger

from .combinatorial_class import CombinatorialClass


class Info(object):
    """Information about a combinatorial class."""
    def __init__(self, comb_class, label, **kwargs):
        """Initialise information."""
        self.comb_class = comb_class
        self.label = label
        self.expanded = kwargs.get('expanded', 0)
        self.symmetry_expanded = kwargs.get('symmetry_expanded', False)
        self.initial_expanded = kwargs.get('initial_expanded', False)
        self.expanding_children_only = kwargs.get('expanding_children_only',
                                                  False)
        self.expanding_other_sym = kwargs.get('expanding_other_sym', False)
        self.expandable = kwargs.get('expandable', False)
        self.inferrable = kwargs.get('inferrable', True)
        self.inferral_expanded = kwargs.get('inferral_expanded', False)
        self.verified = kwargs.get('verified', None)
        self.verification_reason = kwargs.get('verification_reason', None)
        self.empty = kwargs.get('empty', None)
        self.strategy_verified = kwargs.get('strategy_verified', None)

    def to_dict(self):
        """Return dictionary object of self that is JSON serializable."""
        try:
            b64encode(self.comb_class).decode()
        except Exception as e:
            logger.warn(("Lost information about tiling with encoding as:\n"
                         "" + str(self.comb_class) + "\n" + str(e)))
            return None
        return {
            'comb_class': b64encode(self.comb_class).decode(),
            'label': self.label,
            'expanded': self.expanded,
            'symmetry_expanded': self.symmetry_expanded,
            'initial_expanded': self.initial_expanded,
            'expanding_children_only': self.expanding_children_only,
            'expanding_other_sym': self.expanding_other_sym,
            'expandable': self.expandable,
            'inferral_expanded': self.inferral_expanded,
            'verified': self.verified,
            'empty': self.empty,
            'strategy_verified': self.strategy_verified,
        }

    @classmethod
    def from_dict(cls, dict):
        """Return Info object from dictionary."""
        return cls(
            comb_class=b64decode(dict['comb_class'].encode()),
            label=int(dict['label']),
            expanded=int(dict.get('expanded', 0)),
            symmetry_expanded=dict.get('symmetry_expanded', False),
            initial_expanded=dict.get('initial_expanded', False),
            expanding_children_only=dict.get('expanding_children_only', False),
            expanding_other_sym=dict.get('expanding_other_sym', False),
            expandable=dict.get('expandable', False),
            inferral_expanded=dict.get('inferral_expanded', False),
            verified=dict.get('verified', None),
            empty=dict.get('empty', None),
            strategy_verified=dict.get('strategy_verified', False),
        )

    def __eq__(self, other):
        """Equal if all parameters are equal."""
        return (
            self.comb_class == self.comb_class and
            self.label == self.label and
            self.expanded == self.expanded and
            self.symmetry_expanded == self.symmetry_expanded and
            self.initial_expanded == self.initial_expanded and
            self.expanding_children_only == self.expanding_children_only and
            self.expanding_other_sym == self.expanding_other_sym and
            self.expandable == self.expandable and
            self.inferral_expanded == self.inferral_expanded and
            self.verified == self.verified and
            self.empty == self.empty and
            self.strategy_verified == self.strategy_verified
        )


class ClassDB(object):
    """
    A database for combinatorial classes.

    Each class is given a unique integer label. The key to the database is
    both the unique integer label and the ckass. It supports the following
    methods.
    - DB.add(class) will label the class and add it to the database.
    - DB.get_class(key) return the class with the given key.
    - DB.get_label(key) return the label of the given key.
    - DB.set_property(key) will set the property to true for key.
    - DB.is_property(key) will return True if the key has the property, False
    otherwise.
    - Sets verified classes with explanation.
    """

    def __init__(self, combinatorial_class):
        """
        Initialise.

        Two dictionaries allow you to call database with either the
        combinatorial class, or the unique label of the combinatorial class.
        The key can therefore be either the label or the combinatorial class.
        """
        self.class_to_info = {}
        self.label_to_info = {}
        if combinatorial_class is None:
            raise TypeError("Need to declare type of combinatorial class.")
        self.combinatorial_class = combinatorial_class

    def __iter__(self):
        """Iterator of labels."""
        for key in self.label_to_info.keys():
            yield key

    def __contains__(self, key):
        """Check for containment."""
        if isinstance(key, CombinatorialClass):
            self._compress(key)
            info = self.class_to_info.get(key)
        if isinstance(key, int):
            info = self.label_to_info.get(key)
        return info is not None

    def __eq__(self, other):
        """Equal if all information stored is the same."""
        return (self.class_to_info == other.class_to_info and
                self.label_to_info == other.label_to_info)

    def to_dict(self):
        """Return dictionary object of self."""
        return {label: info.to_dict()
                for label, info in self.label_to_info.items()}

    @classmethod
    def from_dict(cls, dict, combinatorial_class):
        """Return ClassDB for dictionary object."""
        classdb = cls(combinatorial_class)
        for label, info in dict.items():
            info = Info.from_dict(info)
            label = info.label
            comb_class = info.comb_class
            classdb.label_to_info[label] = info
            classdb.class_to_info[comb_class] = info
        return classdb

    def add(self,
            comb_class,
            symmetry_expanded=False,
            expanding_other_sym=False,
            expandable=False,
            compressed=False):
        """
        Add a combinatorial class to the database.

        Can also set some information about the combinatorial class on adding.
        """
        if not compressed and not isinstance(comb_class, CombinatorialClass):
            raise TypeError(("Trying to add something that isn't a"
                            "CombinatorialClass."))
        if not compressed:
            compressed_class = self._compress(comb_class)
        else:
            compressed_class = comb_class
        if compressed_class not in self.class_to_info:
            label = len(self.class_to_info)
            info = Info(compressed_class,
                        label,
                        symmetry_expanded=symmetry_expanded,
                        expanding_other_sym=expanding_other_sym,
                        expandable=expandable)
            self.class_to_info[compressed_class] = info
            self.label_to_info[label] = info
        else:
            label = self.class_to_info[compressed_class].label
            if expandable:
                self.set_expandable(label)
            if expanding_other_sym:
                self.set_expanding_other_sym(label)
            if symmetry_expanded:
                self.set_symmetry_expanded(label)

    def _get_info(self, key):
        """Return Info for given key."""
        if isinstance(key, CombinatorialClass):
            key = self._compress(key)
            info = self.class_to_info.get(key)
            if info is None:
                self.add(key, compressed=True)
                info = self.class_to_info[key]
        elif isinstance(key, int):
            info = self.label_to_info.get(key)
            if info is None:
                raise KeyError("Key not in ClassDB.")
        else:
            raise TypeError('ClassDB only accepts'
                            'CombinatorialClass and will decompress with'
                            '{}.'.format(self.combinatorial_class))
        return info

    def _compress(self, key):
        """Return compressed version of combinatorial class."""
        try:
            return key.compress()
        except AttributeError:
            raise AttributeError(("The class {} needs a compress function"
                                  ".").format(self.combinatorial_class))

    def _decompress(self, key):
        """Return decompressed version of compressed combinatorial class."""
        try:
            return self.combinatorial_class.decompress(key)
        except AttributeError:
            raise AttributeError(("The class {} needs a compress function"
                                  ".").format(self.combinatorial_class))

    def get_class(self, key):
        """Return combinatorial class of key."""
        info = self._get_info(key)
        return self._decompress(info.comb_class)

    def get_label(self, key):
        """Return label of key."""
        info = self._get_info(key)
        return info.label

    def number_times_expanded(self, key):
        """
        Return number of times combinatorial class has been expanded.

        i.e., the number of sets strategies that have been used to expanded
        the combinatorial class corresponding to the key.
        """
        info = self._get_info(key)
        return info.expanded

    def increment_expanded(self, key):
        """Increment counter for times combinatorial class was expanded."""
        info = self._get_info(key)
        info.expanded += 1

    def is_expandable(self, key):
        """Return True if expandable, False otherwise."""
        info = self._get_info(key)
        return info.expandable

    def set_expandable(self, key, expandable=True):
        """Update database about combinatorial class inferrability."""
        info = self._get_info(key)
        info.expandable = expandable or bool(info.expandable)

    def is_inferrable(self, key):
        """Return True if inferrable, False otherwise."""
        info = self._get_info(key)
        return info.inferrable

    def set_inferrable(self, key, inferrable=True):
        """Update database about combinatorial class inferrability."""
        info = self._get_info(key)
        info.inferrable = inferrable and bool(info.inferrable)

    def is_verified(self, key):
        """
        Return True if combinatorial class has been verified, and False if not
        verified. Return None if status not set.

        Verification must have an explanation, i.e, from a strategy.
        """
        return self._get_info(key).verified

    def set_verified(self, key, verified=True, explanation=None):
        """
        Update database about combinatorial class being verified

        An explanation should be provided.
        """
        info = self._get_info(key)
        info.verified = verified or bool(info.verified)
        info.verification_reason = explanation

    def verification_reason(self, key):
        """Return explanation of verification, None if not verified."""
        return self._get_info(key).verification_reason

    def is_empty(self, key):
        """Return True if combinatorial class is empty set, False if not.
        Return None if status not set."""
        return self._get_info(key).empty

    def set_empty(self, key, empty=True):
        """Update database about comb class being empty."""
        info = self._get_info(key)
        info.empty = (empty or bool(info.empty))

    def verified_labels(self):
        """Yield all the labels that are verified."""
        for x in self.label_to_info:
            if self.is_verified(x):
                yield x

    def empty_labels(self):
        """Yield all the labels that are verified."""
        for x in self.label_to_info:
            if self.is_empty(x):
                yield x

    def is_strategy_verified(self, key):
        """Return True if combinatorial class verified by a strategy.
        Returns None if never updated."""
        return self._get_info(key).strategy_verified

    def set_strategy_verified(self, key, strategy_verified=True):
        """Update database combinatorial class is verified by a strategy."""
        info = self._get_info(key)
        info.strategy_verified = (strategy_verified or
                                  bool(info.strategy_verified))

    def is_symmetry_expanded(self, key):
        """Return True if combinatorial class was expanded by symmetries."""
        return self._get_info(key).symmetry_expanded

    def set_symmetry_expanded(self, key, symmetry_expanded=True):
        """Update database that combinatorial class was symmetry expanded."""
        info = self._get_info(key)
        info.symmetry_expanded = (symmetry_expanded or
                                  bool(info.symmetry_expanded))

    def is_expanding_children_only(self, key):
        """Return True if not expanding as expanding other children only."""
        return self._get_info(key).expanding_children_only

    def set_expanding_children_only(self, key, expanding_children_only=True):
        """Update database if expanding childern_only."""
        info = self._get_info(key)
        info.expanding_children_only = (expanding_children_only or
                                        bool(info.expanding_children_only))

    def is_expanding_other_sym(self, key):
        """Return True if symmetry of combinatorial class is being expaned."""
        return self._get_info(key).expanding_other_sym

    def set_expanding_other_sym(self, key, expanding_other_sym=True):
        """Update database symmetry of combinatorial class is being
        expanded."""
        info = self._get_info(key)
        info.expanding_other_sym = (expanding_other_sym or
                                    bool(info.expanding_other_sym))

    def is_initial_expanded(self, key):
        """Return True if combinatorial class was initial expanded."""
        return self._get_info(key).initial_expanded

    def set_initial_expanded(self, key, initial_expanded=True):
        """Update database that combinatorial class was initial expanded."""
        info = self._get_info(key)
        info.initial_expanded = (initial_expanded or
                                 bool(info.initial_expanded))

    def is_inferral_expanded(self, key):
        """Return True if combinatorial class was inferral expanded."""
        return self._get_info(key).inferral_expanded

    def set_inferral_expanded(self, key, inferral_expanded=True):
        """Update database that combinatorial class was inferral expanded."""
        info = self._get_info(key)
        info.inferral_expanded = (inferral_expanded or
                                  bool(info.inferral_expanded))



from functools import partial

import sympy
from logzero import logger


def get_func_name(f, warn=False,
                  logger_kwargs={'processname': 'utils'}):
    """Return a string that is the name of the function f."""
    if not callable(f):
        raise TypeError("Function given is not callable.")
    if isinstance(f, partial):
        f = f.func
        if warn:
            logger.warn(("forgetting kwargs given to partial function "
                         "using strategy" + f.__name__),
                        extra=logger_kwargs)
    return f.__name__


def get_module_and_func_names(f, warn=False,
                              logger_kwargs={'processname': 'utils'}):
    """Return the module and name of a function f."""
    if isinstance(f, partial):
        f = f.func
        if warn:
            logger.warn(("forgetting kwargs given to partial function "
                         "using strategy" + f.__name__),
                        extra=logger_kwargs)
    return (f.__module__, f.__name__)


def get_func(module_name, func_name, warn=False,
             logger_kwargs={'processname': 'utils'}):
    """Return the function with given module and function names."""
    import importlib
    try:
        module = importlib.import_module(module_name)
        try:
            func = getattr(module, func_name)
            assert callable(func)
            return func
        except Exception:
            if warn:
                logger.warn(("No function named " + func_name +
                            " in module " + module_name),
                            extra=logger_kwargs)
    except Exception as e:
        if warn:
            logger.warn("No module named " + module_name,
                        extra=logger_kwargs)


def check_poly(min_poly, initial, root_initial=None, root_func=None):
    """Return True if this is a minimum polynomial for the generating
    function F with the given initial terms. Input is a polynomial in F,
    and initial terms."""
    F = sympy.Symbol("F")
    x = sympy.abc.x
    init_poly = 0
    for i, coeff in enumerate(initial):
        init_poly += coeff * x**i
    if root_initial is not None:
        root_poly = 0
        for i, coeff in enumerate(root_initial):
            root_poly += coeff * x**i
    verification = min_poly.subs({F: init_poly})
    if root_initial is not None:
        verification = verification.subs({root_func: root_poly})
    verification = verification.expand()
    verification = verification.series(x, n=len(initial)).removeO()
    verification = (verification +
                    sympy.O(sympy.abc.x**(len(initial) - 1))).removeO()
    return verification == 0


def check_equation(equation, initial, root_initial=None, root_func=None):
    """Return True if an equation in terms of the generating function F and x
    is satisfied."""
    F = sympy.Symbol("F")
    solutions = sympy.solve(equation, F, dict=True,
                            cubics=False, quartics=False, quintics=False)
    for solution in solutions:
        genf = solution[F]
        try:
            expansion = taylor_expand(genf, len(initial) - 1)
        except Exception:
            continue
        if initial == expansion:
            return True
    return False


def get_solution(equation, initial):
    """Return solution of equation in F and x with the given initial
    conditions."""
    F = sympy.Symbol("F")
    solutions = sympy.solve(equation, F, dict=True,
                            cubics=False, quartics=False, quintics=False)
    for solution in solutions:
        genf = solution[F]
        try:
            expansion = taylor_expand(genf, len(initial) - 1)
        except Exception as e:
            continue
        if initial == expansion:
            return genf


def taylor_expand(genf, n=10):
    num, den = genf.as_numer_denom()
    num = num.expand()
    den = den.expand()
    genf = num/den
    ser = sympy.Poly(genf.series(n=n+1).removeO(), sympy.abc.x)
    res = ser.all_coeffs()
    res = res[::-1] + [0]*(n+1-len(res))
    return res


def maple_equations(root_func, root_class, eqs):
    s = "The system of {} equations\n".format(len(eqs))
    s += "root_func := {}:\n".format(str(root_func)).replace("(x)", "")
    s += "eqs := [\n"
    s += ",\n".join("{} = {}".format(str(eq.lhs), str(eq.rhs))
                    for eq in eqs).replace("(x)", "")
    s += "\n]:\n"
    s += "count := {}:".format([len(list(root_class.objects_of_length(i)))
                                for i in range(6)])
    return s


def compositions(n, k):
    # Credit to:
    # https://pythonhosted.org/combalg-py/_modules/combalg/combalg.html
    """
    Iterator over all the composition of n into k parts.

    Composition of are tuple of k integers that are greater of equals to 0 such
    that the sum of the k integers is n.

    >>> sorted(compositions(2,3))
    [(0, 0, 2), (0, 1, 1), (0, 2, 0), (1, 0, 1), (1, 1, 0), (2, 0, 0)]
    >>> sorted(compositions(2,2))
    [(0, 2), (1, 1), (2, 0)]
    >>> list(compositions(2, 1))
    [(2,)]
    >>> list(compositions(0, 1))
    [(0,)]
    >>> list(compositions(-1, 1))
    []
    >>> list(compositions(1, -1))
    []
    >>> list(compositions(1, -1))
    []
    """
    if n < 0 or k < 0 or k == 0:
        return
    t = n
    h = 0
    a = [0]*k
    a[0] = n
    yield tuple(a)
    while a[k-1] != n:
        if t != 1:
            h = 0
        t = a[h]
        a[h] = 0
        a[0] = t-1
        a[h+1] += 1
        h += 1
        yield tuple(a)



"""
A database for rules.

Use to keep track of all batch rules made by strategies. Each rule comes with
an explanantion.
"""
from collections import Iterable, defaultdict


class RuleDB(object):
    """A database for rules found."""

    def __init__(self):
        """
        Initialise.

        - The rules dict keep all rules, where keys are start and items are
        sets of ends.
        - The explanations give reason/formal steps for rules. Call for an
        explanation of rule start->ends with d[start][ends].
        - Some strategies require back maps, these are stored in the back maps
        dictionary. Calling works the same way as explanations.
        """
        self.rules_dict = defaultdict(set)
        self.explanations = {}
        self.constructors = {}

    def __eq__(self, other):
        """Check if all stored information is the same."""
        return (self.rules_dict == other.rules_dict and
                self.explanations == other.explanations and
                self.constructors == other.constructors)

    def to_dict(self):
        """Return dictionary object of self that is JSON serializable."""
        return {
            'rules_dict': [[x, [list(z) for z in y]]
                           for x, y in self.rules_dict.items()],
            'explanations': [[x, [[list(y), z] for y, z in d.items()]]
                             for x, d in self.explanations.items()],
            'constructors': [[x, [[list(y), z] for y, z in d.items()]]
                             for x, d in self.constructors.items()],
        }

    @classmethod
    def from_dict(cls, dict):
        """Return RuleDB object from dictionary."""
        ruledb = RuleDB()
        ruledb.rules_dict = defaultdict(set, {x: {tuple(y) for y in z}
                                        for x, z in dict['rules_dict']})
        ruledb.explanations = {x: {tuple(y): z for y, z in d}
                               for x, d in dict['explanations']}
        ruledb.constructors = {x: {tuple(y): z for y, z in d}
                               for x, d in dict['constructors']}
        return ruledb

    def add(self, start, end, explanation, constructor):
        """
        Add a rule to the database.

        - start is a single integer.
        - end is a tuple of integers.
        - explanation is a string which describes the rule (formal step).
        """
        if not isinstance(start, int):
            raise TypeError("Rule is integer and an iterable of integers.")
        if not isinstance(end, Iterable):
            raise TypeError("Rule is integer and an iterable of integers.")
        if any(not isinstance(x, int) for x in end):
            raise TypeError("Rule is integer and an iterable of integers.")
        if not isinstance(explanation, str):
            raise TypeError("A rule requires a string for an explanation.")
        if not isinstance(constructor, str):
            raise TypeError("A rule requires a string for a constructor.")
            if constructor not in ['cartesian', 'disjoint']:
                raise ValueError("Only handles cartesian and disjoint.")
        end = tuple(sorted(end))
        self.rules_dict[start] |= set((end,))
        if start in self.explanations:
            self.explanations[start][end] = explanation
        else:
            self.explanations[start] = {end: explanation}
        if start in self.constructors:
            self.constructors[start][end] = constructor
        else:
            self.constructors[start] = {end: constructor}

    def remove(self, start, end):
        """Remove rule from database."""
        if not isinstance(start, int):
            raise TypeError("Rule is integer and an iterable of integers.")
        if not isinstance(end, Iterable):
            raise TypeError("Rule is integer and an iterable of integers.")
        if any(not isinstance(x, int) for x in end):
            raise TypeError("Rule is integer and an iterable of integers.")
        if start in self.explanations:
            if end in self.explanations[start]:
                self.explanations[start].pop(end)
                if not self.explanations[start]:
                    self.explanations.pop(start)
        if start in self.constructors:
            if end in self.constructors[start]:
                self.constructors[start].pop(end)
                if not self.constructors[start]:
                    self.constructors.pop(start)
        if start in self.rules_dict:
            if end in self.rules_dict[start]:
                self.rules_dict[start].remove(end)
                if not self.rules_dict[start]:
                    self.rules_dict.pop(start)

    def __iter__(self):
        """Iterate through rules as the pairs (start, end)."""
        for start, ends in self.rules_dict.items():
            for end in ends:
                yield start, end

    def explanation(self, start, end):
        """Return the explanation of the rule start -> end."""
        end = tuple(sorted(end))
        if (start not in self.explanations or
                end not in self.explanations[start]):
            raise KeyError("No such rule.")
        return self.explanations[start][end]

    def constructor(self, start, end):
        """Return the constructor of the rule start -> end."""
        end = tuple(sorted(end))
        if (start not in self.constructors or
                end not in self.constructors[start]):
            raise KeyError("No such rule.")
        return self.constructors[start][end]

    def contains(self, start, end):
        """Return true if the rule start -> end is in the database."""
        end = tuple(sorted(end))
        return start in self.rules_dict and end in self.rules_dict[start]



from .rule import (BatchRule, BatchStrategy, DecompositionRule,
                   DecompositionStrategy, EquivalenceRule, EquivalenceStrategy,
                   InferralRule, InferralStrategy, Rule, Strategy,
                   VerificationRule, VerificationStrategy)
from .strategy_pack import StrategyPack



"""
A queue of labels.
"""
import sys
from collections import deque


class ClassQueue(object):
    """
    The Queue determines the order that classes are expanded.
    """
    def __init__(self):
        """Initialise a ClassQueue."""
        self.working = deque()
        self.curr_level = deque()
        self.next_level = deque()
        self.levels_completed = 0
        self.ignore = set()

    def to_dict(self):
        """Return dictionary object of self that is JSON serializable."""
        return {
            'working': list(self.working),
            'curr_level': list(self.curr_level),
            'next_level': list(self.next_level),
            'levels_completed': self.levels_completed,
            'ignore': list(self.ignore),
        }

    @classmethod
    def from_dict(cls, dict):
        """Return ClassQueue from dictionary object."""
        queue = cls()
        queue.working = deque(dict['working'])
        queue.curr_level = deque(dict['curr_level'])
        queue.next_level = deque(dict['next_level'])
        queue.levels_completed = dict['levels_completed']
        queue.ignore = set(dict['ignore'])
        return queue

    def add_to_working(self, comb_class):
        """Add combinatorial class to working queue."""
        self.working.append(comb_class)

    def add_to_next(self, comb_class):
        """Add combinatorial class to next queue."""
        self.next_level.append(comb_class)

    def add_to_curr(self, comb_class):
        """Add combinatorial class to current queue."""
        self.curr_level.append(comb_class)

    def next(self):
        """
        Return next combinatorial class in current queue.

        If current queue becomes empty will change next queue to current and
        return first combinatorial class. Return None if no combinatorial
        classes to expand.
        """
        while True:
            if self.working:
                n = self.working.popleft()
            elif self.curr_level:
                n = self.curr_level.popleft()
            else:
                if not self.next_level:
                    return None
                self.levels_completed += 1
                self.curr_level = self.next_level
                self.ignore = set()
                self.next_level = deque()
                continue
            if n not in self.ignore:
                return n

    def do_level(self):
        """
        An iterator of all combinatorial classes in the current queue.

        Will swap next queue to current after iteration. Yields None if no
        combinatorial class to expand.
        """
        curr_level = self.levels_completed
        while curr_level == self.levels_completed:
            n = self.next()
            yield n
            if n is None:
                return None



import json
import logging
import os
import time
import warnings
from base64 import b64decode, b64encode
from collections import defaultdict
from functools import partial, reduce
from operator import add, mul

import logzero
import psutil
import sympy
from logzero import logger

from comb_spec_searcher.utils import compositions

from .class_db import ClassDB
from .class_queue import ClassQueue
from .equiv_db import EquivalenceDB
from .proof_tree import ProofTree as ProofTree
from .rule_db import RuleDB
from .strategies import Rule, StrategyPack, VerificationRule
from .tree_searcher import (iterative_proof_tree_finder, iterative_prune,
                            proof_tree_generator_bfs, proof_tree_generator_dfs,
                            prune, random_proof_tree)
from .utils import get_func, get_func_name, get_module_and_func_names

warnings.simplefilter("once", Warning)


class CombinatorialSpecificationSearcher(object):
    """
    The CombinatorialSpecificationSearcher classs.

    This is used to build up knowledge about a combinatorial_class with respect
    to the given strategies and search for a combinatorial specification.
    """
    def __init__(self, start_class, strategy_pack, **kwargs):
        """Initialise CombinatorialSpecificationSearcher."""
        if start_class is None:
            raise ValueError(("CombinatorialSpecificationSearcher"
                              "requires a start class."))
        self.start_class = start_class
        if strategy_pack is None:
            raise ValueError(("CombinatorialSpecificationSearcher"
                              "requires a strategy pack."))
        else:
            if not isinstance(strategy_pack, StrategyPack):
                raise TypeError(("Strategy pack given not "
                                 "instance of strategy pack."))
            self.strategy_pack = strategy_pack
        self.debug = kwargs.get('debug', False)
        if not self.debug:
            logzero.loglevel(logging.INFO, True)
        self.kwargs = kwargs.get('function_kwargs', dict())
        self.logger_kwargs = kwargs.get('logger_kwargs',
                                        {'processname': 'runner'})
        self.kwargs['logger'] = self.logger_kwargs

        self.initial_strategies = strategy_pack.initial_strats
        self.strategy_generators = strategy_pack.expansion_strats
        self.inferral_strategies = strategy_pack.inferral_strats
        self.verification_strategies = strategy_pack.ver_strats
        self.iterative = (strategy_pack.iterative or
                          kwargs.get('iterative', False))
        self.forward_equivalence = (strategy_pack.forward_equivalence or
                                    kwargs.get('forward_equivalence', False))
        self.symmetries = kwargs.get('symmetries', strategy_pack.symmetries)
        if self.symmetries:
            # A list of symmetry functions of classes.
            if (not isinstance(self.symmetries, list) or
                    any(not callable(f) for f in self.symmetries)):
                raise ValueError(("To use symmetries need to give a"
                                  "list of symmetry functions."))
            self.kwargs['symmetry'] = True
        else:
            self.symmetries = []

        self.classdb = ClassDB(type(start_class))
        self.equivdb = EquivalenceDB()
        self.classqueue = ClassQueue()
        self.ruledb = RuleDB()

        self.classdb.add(start_class, expandable=True)
        self.start_label = self.classdb.get_label(start_class)
        self.classqueue.add_to_working(self.start_label)

        self._has_proof_tree = False

        self.strategy_times = defaultdict(int)
        self.strategy_expansions = defaultdict(int)
        self.symmetry_time = 0
        self.tree_search_time = 0
        self.prep_for_tree_search_time = 0
        self.queue_time = 0
        self._time_taken = 0
        self.class_genf = {}

    def to_dict(self):
        return {
            'start_class': b64encode(self.start_class.compress()).decode(),
            'debug': self.debug,
            'kwargs': self.kwargs,
            'logger_kwargs': self.logger_kwargs,
            'initial_strategies': [get_module_and_func_names(
                                f, warn=True, logger_kwargs=self.logger_kwargs)
                                   for f in self.initial_strategies],
            'strategy_generators': [[get_module_and_func_names(
                                f, warn=True, logger_kwargs=self.logger_kwargs)
                                     for f in l]
                                    for l in self.strategy_generators],
            'inferral_strategies': [get_module_and_func_names(
                                f, warn=True, logger_kwargs=self.logger_kwargs)
                                    for f in self.inferral_strategies],
            'verification_strategies': [get_module_and_func_names(
                                f, warn=True, logger_kwargs=self.logger_kwargs)
                                        for f in self.verification_strategies],
            'iterative': self.iterative,
            'forward_equivalence': self.forward_equivalence,
            'symmetries': self.symmetries,
            'classdb': self.classdb.to_dict(),
            'equivdb': self.equivdb.to_dict(),
            'classqueue': self.classqueue.to_dict(),
            'ruledb': self.ruledb.to_dict(),
            'start_label': self.start_label,
            '_has_proof_tree': self._has_proof_tree,
            'strategy_times': dict(self.strategy_times),
            'strategy_expansions': dict(self.strategy_expansions),
            'symmetry_time': self.symmetry_time,
            'tree_search_time': self.tree_search_time,
            'prep_for_tree_search_time': self.prep_for_tree_search_time,
            'queue_time': self.queue_time,
            '_time_taken': self._time_taken,
        }

    @classmethod
    def from_dict(cls, dict, combinatorial_class):
        from comb_spec_searcher.strategies.strategy_pack import StrategyPack
        strategy_pack = StrategyPack(
            initial_strats=[get_func(*mod_func_name)
                            for mod_func_name in dict['initial_strategies']],
            ver_strats=[get_func(*mod_func_name)
                        for mod_func_name in dict['verification_strategies']],
            inferral_strats=[get_func(*mod_func_name)
                             for mod_func_name in dict['inferral_strategies']],
            expansion_strats=[[get_func(*mod_func_name)
                               for mod_func_name in l]
                              for l in dict['strategy_generators']],
            name="recovered strategy pack")

        kwargs = {
          'debug': dict['debug'],
          'iterative': dict['iterative'],
          'forward_equivalence': dict['forward_equivalence'],
        }

        try:
            kwargs['logger_kwargs'] = dict['logger_kwargs']
        except Exception:
            logger.warn('logger_kwargs could not be recovered')
        try:
            kwargs['function_kwargs'] = dict['function_kwargs']
        except Exception:
            logger.warn('function_kwargs could not be recovered')
        b = b64decode(dict['start_class'].encode())
        c = combinatorial_class.decompress(b)
        css = cls(c, strategy_pack, **kwargs)
        css.classdb = ClassDB.from_dict(dict['classdb'],
                                        combinatorial_class)
        css.equivdb = EquivalenceDB.from_dict(dict['equivdb'])
        css.classqueue = ClassQueue.from_dict(dict['classqueue'])
        css.ruledb = RuleDB.from_dict(dict['ruledb'])
        css.start_label = dict['start_label']
        css._has_proof_tree = dict['_has_proof_tree']
        css.strategy_times = defaultdict(int, dict['strategy_times'])
        css.strategy_expansions = defaultdict(int, dict['strategy_expansions'])
        css.symmetry_time = dict['symmetry_time']
        css.tree_search_time = dict['tree_search_time']
        css.prep_for_tree_search_time = dict['prep_for_tree_search_time']
        css.queue_time = dict['queue_time']
        css._time_taken = dict['_time_taken']
        return css

    def update_status(self, strategy, time_taken):
        """Update that it took 'time_taken' to expand a combinatorial class
        with strategy"""
        name = get_func_name(strategy)
        self.strategy_times[name] += time_taken
        self.strategy_expansions[name] += 1

    def try_verify(self, comb_class, label, force=False):
        """
        Try to verify the combinatorial class.

        It will only try to verify combinatorial classes who have no equivalent
        combinatorial classes already verified. If force=True, it will try to
        verify combinatorial class if it is not already strategy verified.
        """
        if force:
            if self.classdb.is_strategy_verified(label):
                return
        elif self.equivdb.is_verified(label):
            return
        if self.classdb.is_strategy_verified(label) is None:
            for ver_strategy in self.verification_strategies:
                start = time.time()
                rule = ver_strategy(comb_class, **self.kwargs)
                self.update_status(ver_strategy, time.time() - start)
                if rule is not None:
                    if not isinstance(rule, VerificationRule):
                        raise TypeError(("Attempting to verify with non "
                                         "VerificationRule."))
                    formal_step = rule.formal_step
                    self.classdb.set_verified(label, explanation=formal_step)
                    self.classdb.set_strategy_verified(label)
                    self.equivdb.update_verified(label)
                    return
            self.classdb.set_strategy_verified(label, False)

    def is_empty(self, comb_class, label):
        """Return True if a combinatorial class contains no objects, False
        otherwise."""
        start = time.time()
        if self.classdb.is_empty(label) is None:
            if comb_class.is_empty():
                self._add_empty_rule(label)
                assert self.classdb.is_empty(label)
            else:
                self.classdb.set_empty(label, empty=False)
                assert not self.classdb.is_empty(label)
            self.update_status(self.is_empty, time.time() - start)
        return self.classdb.is_empty(label)

    def _symmetric_classes(self, comb_class, explanation=False):
        """Return all symmetries of a combinatorial class.

        This function only works if symmetry strategies have been given to the
        CombinatorialSpecificationSearcher.
        """
        symmetric_classes = []
        for sym in self.symmetries:
            symmetric_class = sym(comb_class)
            if symmetric_class != comb_class:
                if explanation:
                    if all(x != symmetric_class
                           for x, _ in symmetric_classes):
                        symmetric_classes.append((symmetric_class,
                                                  str(sym).split(' ')[1]))
                else:
                    if all(x != symmetric_class
                           for x in symmetric_classes):
                        symmetric_class.append(symmetric_class)
        return symmetric_classes

    def expand(self, label):
        """
        Will expand the combinatorial class with given label.

        The first time function called with label, it will expand with the
        inferral strategies. The second time with the initial strategies.  The
        third time the first set of other strategies, fourth time the second
        set etc.
        """
        start = time.time()
        comb_class = self.classdb.get_class(label)
        self.queue_time += time.time() - start

        if (self.classdb.is_inferrable(label) and
                not self.classdb.is_inferral_expanded(label)):
            logger.debug('Inferring label {}'.format(str(label)),
                         extra=self.logger_kwargs)
            self._inferral_expand(comb_class, label)
            self.classqueue.add_to_working(label)
        elif self.classdb.is_expandable(label):
            if not self.classdb.is_initial_expanded(label):
                logger.debug('Initial expanding label {}'.format(str(label)),
                             extra=self.logger_kwargs)
                self._initial_expand(comb_class, label)
                self.classqueue.add_to_next(label)
            else:
                expanding = self.classdb.number_times_expanded(label)
                logger.debug('Expanding label {}'.format(str(label)),
                             extra=self.logger_kwargs)
                strategies = self.strategy_generators[expanding]
                for strategy_generator in strategies:
                    # function returns time it took.
                    self._expand_class_with_strategy(comb_class,
                                                     strategy_generator, label)

                if not self.is_expanded(label):
                    self.classdb.increment_expanded(label)
                    if not self.is_expanded(label):
                        self.classqueue.add_to_curr(label)

    def _expand_class_with_strategy(self, comb_class, strategy_function,
                                    label, initial=False, inferral=False):
        """
        Will expand the class with given strategy. Return time taken.

        If 'inferral' will return time also return inferred class and label.
        """
        start = time.time()
        if inferral:
            strat = strategy_function(comb_class, **self.kwargs)
            inf_class = None
            inf_label = None
            if strat is None:
                strategy_generator = []
            else:
                strategy_generator = [strat]
        else:
            strategy_generator = strategy_function(comb_class, **self.kwargs)
        for rule in strategy_generator:
            if not isinstance(rule, Rule):
                raise TypeError("Attempting to add non Rule type.")
            if inferral and len(rule.comb_classes) != 1:
                raise TypeError(("Attempting to infer with non "
                                 "inferral strategy."))
            if inferral and comb_class == rule.comb_classes[0]:
                logger.debug(("The inferral strategy {} returned the same "
                              "combinatorial class when applied to {}"
                              "".format(str(rule).split(' ')[1],
                                        repr(comb_class))),
                             extra=self.logger_kwargs)
                continue
            labels = [self.classdb.get_label(comb_class)
                      for comb_class in rule.comb_classes]
            logger.debug(("Adding combinatorial rule {} -> {} with constructor"
                          " '{}'".format(label, tuple(labels),
                                         rule.constructor)),
                         extra=self.logger_kwargs)

            if any(self.equivdb[label] == self.equivdb[l] for l in labels):
                # This says comb_class = comb_class, so we skip it, but mark
                # every other class as empty.
                for l in labels:
                    if self.equivdb[label] != self.equivdb[l]:
                        self._add_empty_rule(l)
                if self.debug:
                    for l, c in zip(labels, rule.comb_classes):
                        if self.equivdb[label] != self.equivdb[l]:
                            assert c.is_empty()

            start -= time.time()
            end_labels, classes, formal_step = self._rule_cleanup(rule,
                                                                  labels)
            start += time.time()

            if rule.ignore_parent:
                if all(self.classdb.is_expandable(x) for x in end_labels):
                    self.classdb.set_expanding_children_only(label)

            if not end_labels:
                # all the classes are empty so the class itself must be empty!
                self._add_empty_rule(label, "batch empty")
                break
            elif len(end_labels) == 1:
                # If we have an equivalent rule
                self._add_equivalent_rule(label, end_labels[0], formal_step,
                                          rule.constructor)
                if inferral:
                    inf_class = classes[0]
                    inf_label = end_labels[0]
            else:
                constructor = rule.constructor
                self._add_rule(label, end_labels, formal_step, constructor)

            for end_label in end_labels:
                self._add_to_queue(end_label, initial, inferral)

        self.update_status(strategy_function, time.time() - start)

        if inferral:
            return inf_class, inf_label

    def _add_equivalent_rule(self, start, end, explanation, constructor):
        """Add equivalent rule to equivdb and equivalent combinatorial
        class to queue"""
        if explanation is None:
            explanation = "They are equivalent."
        if start == end:
            logger.debug(("Skipping adding equivalent rule with identical"
                          " combinatorial classes."), extra=self.logger_kwargs)
            return
        if self.debug:
            try:
                self._sanity_check_rule(start, [end], 'equiv')
            except Exception:
                error = ("Equivalent rule did not work\n" +
                         repr(self.classdb.get_class(start)) + "\n" +
                         "is not equivalent to" + "\n" +
                         repr(self.classdb.get_class(end)) + "\n" +
                         "formal step:" + explanation)
                logger.debug(error, extra=self.logger_kwargs)
        if self.forward_equivalence:
            reverse_rule = end, (start,)
            if self.ruledb.contains(*reverse_rule):
                raise ValueError(("Same equivalent rule found forward and "
                                  "backwards."))
            else:
                self._add_rule(start, [end], explanation, constructor)
        else:
            self.equivdb.union(start, end, explanation)

    def _add_rule(self, start, ends, explanation=None, constructor=None):
        """Add rule to the rule database and end labels to queue."""
        if explanation is None:
            explanation = "Some rule."
        if constructor is None:
            logger.debug("Assuming constructor is disjoint.",
                         extra=self.logger_kwargs)
            constructor = 'disjoint'
        if self.debug:
            try:
                self._sanity_check_rule(start, ends, constructor)
            except Exception:
                error = ("Expansion rule did not work\n" +
                         repr(self.classdb.get_class(start)) + "\n" +
                         "is equivalent to" + "\n" +
                         repr([self.classdb.get_class(e) for e in ends]) +
                         "\nformal step:" + explanation)
                logger.debug(error, extra=self.logger_kwargs)
        self.ruledb.add(start,
                        ends,
                        explanation,
                        constructor)

    def _add_to_queue(self, label, initial=False, inferral=False):
        """Add a label back onto the queue."""
        if inferral or not initial:
            self.classqueue.add_to_working(label)
        else:
            self.classqueue.add_to_next(label)

    def _add_empty_rule(self, label, explanation=None):
        """Mark label as empty. Treated as verified as can count empty set."""
        if self.classdb.is_empty(label):
            return
        self.classdb.set_empty(label)
        self.classdb.set_verified(label,
                                  explanation="Contains no avoiding objects.")
        self.classdb.set_strategy_verified(label)
        self.equivdb.update_verified(label)

    def _sanity_check_rule(self, start, ends, constructor, length=5):
        """
        Sanity check a rule that has been found up to the length given.
        (default: length=5)
        """
        start_class = self.classdb.get_class(start)
        end_classes = [self.classdb.get_class(e) for e in ends]
        start_count = [len(list(start_class.objects_of_length(i)))
                       for i in range(length)]
        end_counts = [[len(list(e.objects_of_length(i)))
                       for i in range(length)] for e in end_classes]

        if constructor == 'equiv':
            assert len(end_classes) == 1
            assert start_count == end_counts[0]
        elif constructor == 'disjoint':
            assert start_count == [sum(c[i] for c in end_counts)
                                   for i in range(length)]
        elif constructor == 'cartesian':
            for i in range(length):
                total = 0
                for part in compositions(i, len(end_counts)):
                    subtotal = 1
                    for end_count, partlen in zip(end_counts, part):
                        if subtotal == 0:
                            break
                        subtotal *= end_count[partlen]
                    total += subtotal
                assert total == start_count[i]

    def _rule_cleanup(self, rule, labels):
        """
        Return the cleaned rules labels, classes and updated formal step.

        - infer combinatorial classes
        - try to verify combinatorial classes
        - set workability of combinatorial classes
        - remove empty combinatorial classes
        - symmetry expand combinatorial classes
        """
        end_labels = []
        end_classes = []
        inferral_steps = []
        for comb_class, infer, pos_empty, work, label in zip(
                                                        rule.comb_classes,
                                                        rule.inferable,
                                                        rule.possibly_empty,
                                                        rule.workable, labels):
            inferral_step = ""
            if self.symmetries:
                self._symmetry_expand(comb_class)

            if infer:
                self.classdb.set_inferrable(label)
            if work:
                self.classdb.set_expandable(label)
            if infer or work:
                self.classqueue.add_to_working(label)

            # Only applying is_empty check to inferrable comb classes that are
            # possibly empty.
            if infer and pos_empty and self.is_empty(comb_class, label):
                logger.debug("Label {} is empty.".format(label),
                             extra=self.logger_kwargs)
                inferral_steps.append(inferral_step + "Class is empty.")
                continue
            if not pos_empty:
                self.classdb.set_empty(label, empty=False)

            self.try_verify(comb_class, label)

            end_classes.append(comb_class)
            end_labels.append(label)
            inferral_steps.append(inferral_step)

        inferral_step = "~"
        for i, s in enumerate(inferral_steps):
            inferral_step = inferral_step + "[" + str(i) + ": " + s + "]"
        inferral_step = inferral_step + "~"
        formal_step = rule.formal_step + inferral_step

        return end_labels, end_classes, formal_step

    def is_expanded(self, label):
        """Return True if a combinatorial class has been expanded by all
        strategies."""
        number_times_expanded = self.classdb.number_times_expanded(label)
        if number_times_expanded >= len(self.strategy_generators):
            return True
        return False

    def _symmetry_expand(self, comb_class):
        """Add symmetries of combinatorial class to the database."""
        start = time.time()
        if not self.classdb.is_symmetry_expanded(comb_class):
            for sym_comb_class, formal_step in self._symmetric_classes(
                                                comb_class, explanation=True):
                self.classdb.add(sym_comb_class,
                                 expanding_other_sym=True,
                                 symmetry_expanded=True)
                self.equivdb.union(self.classdb.get_label(comb_class),
                                   self.classdb.get_label(sym_comb_class),
                                   formal_step)
        self.symmetry_time += time.time() - start

    def _inferral_expand(self, comb_class, label, inferral_strategies=None,
                         skip=None, start_index=None):
        """
        Inferral expand combinatorial class with given label and inferral
        strategies.

        It will apply all inferral strategies to an combinatorial class.
        Return True if combinatorial class is inferred.
        """
        if self.debug:
            assert comb_class == self.classdb.get_class(label)
        if self.classdb.is_inferral_expanded(label):
            return
        if inferral_strategies is None:
            inferral_strategies = self.inferral_strategies
        for i, strategy_generator in enumerate(inferral_strategies):
            if strategy_generator == skip:
                continue
            inf_class, inf_label = self._expand_class_with_strategy(
                                                comb_class, strategy_generator,
                                                label, inferral=True)
            if inf_class is not None:
                self.classdb.set_inferral_expanded(label)
                inferral_strategies = (inferral_strategies[i + 1:] +
                                       inferral_strategies[0:i + 1])
                self._inferral_expand(inf_class, inf_label,
                                      inferral_strategies,
                                      skip=strategy_generator)
                break
        self.classdb.set_inferral_expanded(label)

    def _initial_expand(self, comb_class, label):
        """
        Expand comb_class with given label using initial strategies.

        It will apply all of the initial strategies.
        """
        if self.debug:
            if comb_class != self.classdb.get_class(label):
                raise ValueError("comb_class and label should match")
        for strategy_generator in self.initial_strategies:
            if (not self.classdb.is_expandable(label) or
                    self.classdb.is_initial_expanded(label) or
                    self.classdb.is_expanding_other_sym(label) or
                    self.classdb.is_expanding_children_only(label)):
                break
            self._expand_class_with_strategy(comb_class,
                                             strategy_generator,
                                             label,
                                             initial=True)

        self.classdb.set_initial_expanded(label)
        self.classqueue.ignore.add(label)

    def get_equations(self, **kwargs):
        """
        Returns a set of equations for all rules currently found.

        If keyword substitutions=True is given, then will return the equations
        with complex functions replaced by symbols and a dictionary of
        substitutions.

        If keyword fake_verify=label it will verify label and return equations
        which are in a proof tree for the root assuming that label is verified.
        """
        if kwargs.get('substitutions'):
            # dictionary from comb_class to symbol, filled by the combinatorial
            # class' get_genf function.
            kwargs['symbols'] = {}
            # dictionary from symbol to genf, filled by the combinatorial
            # class' get_genf function.
            kwargs['subs'] = {}

        functions = {}

        def get_function(label):
            """Return sympy function with label."""
            label = self.equivdb[label]
            function = functions.get(label)
            if function is None:
                function = sympy.Function("F_" + str(label))(sympy.abc.x)
                functions[label] = function
            return function

        if kwargs.get('fake_verify'):
            rules_dict = self.tree_search_prep()
            if kwargs.get('fake_verify'):
                rules_dict[kwargs.get('fake_verify')].add(tuple())
            rules_dict = prune(rules_dict)
            verified_labels = set(rules_dict.keys())
            if kwargs.get('fake_verify'):
                if self.start_label not in verified_labels:
                    return set()
                strat_ver = set()

        equations = set()

        for start, ends in self.ruledb:
            if kwargs.get('fake_verify'):
                if (self.equivdb[start] not in verified_labels or
                    any(self.equivdb[x] not in verified_labels
                        for x in ends)):
                    continue
                strat_ver.add(start)
                strat_ver.update(ends)

            start_function = get_function(start)
            end_functions = (get_function(end) for end in ends)
            constructor = self.ruledb.constructor(start, ends)
            if constructor == 'disjoint' or constructor == 'equiv':
                eq = sympy.Eq(start_function,
                              reduce(add, [f for f in end_functions], 0))
            elif constructor == 'cartesian':
                eq = sympy.Eq(start_function,
                              reduce(mul, [f for f in end_functions], 1))
            else:
                raise NotImplementedError(("Only handle cartesian and "
                                           "disjoint. Don't understand"
                                           " {}.".format(constructor)))
            equations.add(eq)

        kwargs['root_func'] = get_function(self.start_label)
        kwargs['root_class'] = self.start_class
        for label in self.classdb:
            if kwargs.get('fake_verify') and label not in strat_ver:
                continue
            if self.classdb.is_strategy_verified(label):
                try:
                    function = get_function(label)
                    comb_class = self.classdb.get_class(label)
                    gen_func = self.get_class_genf(comb_class, **kwargs)
                    eq = sympy.Eq(function, gen_func)
                    equations.add(eq)
                except Exception as e:
                    error = ("Failed to find generating function for:\n" +
                             repr(comb_class) + "\nVerified as:\n" +
                             self.classdb.verification_reason(label) +
                             "\nThe error was:\n" + str(e))
                    logger.warn(error, extra=self.logger_kwargs)

        if kwargs.get('substitutions'):
            return equations, [sympy.Eq(lhs, rhs)
                               for lhs, rhs in kwargs.get('subs').items()]
        return equations

    def get_class_genf(self, comb_class, **kwargs):
        genf = self.class_genf.get(comb_class)
        if genf is None:
            def taylor_expand(genf, n=10):
                num, den = genf.as_numer_denom()
                num = num.expand()
                den = den.expand()
                genf = num/den
                ser = sympy.Poly(genf.series(n=n+1).removeO(), sympy.abc.x)
                res = ser.all_coeffs()
                res = res[::-1] + [0]*(n+1-len(res))
                return res
            genf = sympy.sympify(comb_class.get_genf(**kwargs))
            if not kwargs['root_func'] in genf.atoms(sympy.Function):
                count = [len(list(comb_class.objects_of_length(i)))
                         for i in range(9)]
                if taylor_expand(sympy.sympify(genf), 8) != count:
                    raise ValueError(("Incorrect generating function "
                                      "in database.\n" + repr(comb_class)))
            self.class_genf[comb_class] = genf
        return genf

    def do_level(self):
        """Expand combinatorial classes in current queue. Combintorial classes
        found added to next."""
        start = time.time()
        queue_start = time.time()
        for label in self.classqueue.do_level():
            if label is None:
                return True
            if self.is_expanded(label):
                continue
            elif self.equivdb.is_verified(label):
                continue
            elif self.classdb.is_expanding_children_only(label):
                continue
            elif not self.classdb.is_expandable(label):
                continue
            elif self.classdb.is_expanding_other_sym(label):
                continue
            queue_start -= time.time()
            self.expand(label)
            queue_start += time.time()
        self.queue_time += time.time() - queue_start
        self._time_taken += time.time() - start

    def expand_classes(self, total):
        """Will send 'total' many classes to the expand function."""
        start = time.time()
        queue_start = time.time()
        count = 0
        while count < total:
            label = self.classqueue.next()
            if label is None:
                return True
            if self.is_expanded(label):
                continue
            elif self.equivdb.is_verified(label):
                continue
            elif self.classdb.is_expanding_children_only(label):
                continue
            elif not self.classdb.is_expandable(label):
                continue
            elif self.classdb.is_expanding_other_sym(label):
                continue
            count += 1
            queue_start -= time.time()
            self.expand(label)
            queue_start += time.time()
        self.queue_time += time.time() - queue_start
        self._time_taken += time.time() - start

    def status(self):
        """
        Return a string of the current status of the CombSpecSearcher.

        It includes:
        - number of combinatorial classes, and information about verification
        - the times spent in each of the main functions
        """
        status = ""
        status += "Currently on 'level' {}\n".format(
                                        self.classqueue.levels_completed + 1)
        status += "Time spent searching so far: ~{} seconds\n".format(
                                                        int(self._time_taken))

        status += "Memory (alone and shared) currently in use: {}\n".format(
                                                        self.get_mem())
        all_labels = self.classdb.label_to_info.keys()
        status += "Total number of combinatorial classes found is {}\n".format(
                                                        str(len(all_labels)))
        expandable = 0
        verified = 0
        strategy_verified = 0
        empty = 0
        equivalent_sets = set()
        verified_equiv_sets = set()
        for label in all_labels:
            if (not self.is_expanded(label) and
                not self.equivdb.is_verified(label) and
                not self.classdb.is_expanding_children_only(label) and
                self.classdb.is_expandable(label) and
                    not self.classdb.is_expanding_other_sym(label)):
                expandable += 1
            if self.equivdb.is_verified(label):
                verified += 1
                verified_equiv_sets.add(self.equivdb[label])
            if self.classdb.is_strategy_verified(label):
                strategy_verified += 1
            if self.classdb.is_empty(label):
                empty += 1
            equivalent_sets.add(self.equivdb[label])

        status += ("Total number of expandable combinatorial classes is {}\n"
                   "".format(expandable))
        status += ("Total number of verified combinatorial classes is {}\n"
                   "".format(verified))
        status += ("Total number of strategy verified combinatorial "
                   "classes is {}\n".format(strategy_verified))
        status += ("Total number of empty combinatorial classes is {}\n"
                   "".format(empty))

        status += ("Total number of equivalent sets is {}\n"
                   "".format(len(equivalent_sets)))
        status += ("Total number of verified equivalent sets is {}\n"
                   "".format(len(verified_equiv_sets)))

        status += ("Total number of combinatorial rules is {}\n"
                   "".format(len(list(self.ruledb))))

        status += "The size of the working queue is {}\n".format(
                                            len(self.classqueue.working))
        status += "The size of the current queue is {}\n".format(
                                         len(self.classqueue.curr_level))
        status += "The size of the next queue is {}\n".format(
                                         len(self.classqueue.next_level))

        for strategy, number in self.strategy_expansions.items():
            status += "Applied {} to {} combinatorial classes\n".format(
                                                        strategy, number)

        symme_perc = 0 if self._time_taken == 0 else \
            int(self.symmetry_time/self._time_taken * 100)
        strat_perc = 0

        for strategy, total_time in self.strategy_times.items():
            perc = 0 if self._time_taken == 0 else \
                total_time/self._time_taken * 100
            strat_perc += perc
            status += "Time spent applying {}: {} seconds, ~{}%\n".format(
                                    strategy, int(total_time), int(perc))

        if self.symmetries:
            status += ("Time spent symmetry applying:"
                       "~{} seconds, ~{}%\n").format(
                        int(self.symmetry_time), int(symme_perc))

        queue_perc = 0 if self._time_taken == 0 else \
            self.queue_time/self._time_taken * 100

        prep_time = self.prep_for_tree_search_time

        prpts_perc = 0 if self._time_taken == 0 else \
            prep_time/self._time_taken * 100

        tsrch_perc = 0 if self._time_taken == 0 else \
            self.tree_search_time/self._time_taken * 100

        status += "Time spent queueing: ~{} seconds, ~{}%\n".format(
                                        int(self.queue_time), int(queue_perc))
        status += ("Time spent prepping for tree search: ~{} seconds, "
                   "~{}%\n").format(int(prep_time), int(prpts_perc))
        status += "Time spent searching for tree: ~{} seconds, ~{}%\n".format(
                                int(self.tree_search_time), int(tsrch_perc))

        total_perc = (strat_perc + symme_perc +
                      queue_perc + prpts_perc + tsrch_perc)
        status += "Total of ~{}% accounted for.\n".format(int(total_perc))
        return status

    def get_mem(self):
        """Return memory used by CombSpecSearcher - note this is actually the
        memory usage of the process that the instance of CombSpecSearcher was
        invoked."""
        mem = psutil.Process(os.getpid()).memory_info().rss
        if (mem / 1024**3 < 1):
            return str(round(mem / 1024**2))+" MiB"
        else:
            return str(round(mem / 1024**3, 3))+" GiB"

    def run_information(self):
        """Return string detailing what CombSpecSearcher is looking for."""
        start_string = ("Initialising CombSpecSearcher for the combinatorial"
                        " class:\n{}\n".format(self.start_class))
        start_string += str(self.strategy_pack)
        return start_string

    def auto_search(self, perc=1, status_update=None, max_time=None,
                    save=False, smallest=False, genf=False):
        """
        An automatic search function.

        It will expand classes until perc*(tree search time) has passed and
        then search for a tree.

        Information is logged to logger.info. A status update is given when a
        tree is found and after status_update many seconds have passed.
        It will also log the proof tree, in json format.

        If save, it will log a json string of CombSpecSearcher to
        logger.info.
        """
        if not 0 < perc <= 100:
            logger.warn(("Percentage not between 0 and 100, so assuming 1%"
                         " search percentage."), extra=self.logger_kwargs)
            perc = 1
        if status_update:
            status_start = time.time()
        start_string = "Auto search started {}\n".format(
                    time.strftime("%a, %d %b %Y %H:%M:%S", time.gmtime()))
        start_string += self.run_information()
        logger.info(start_string, extra=self.logger_kwargs)

        max_search_time = 0
        expanding = True
        while expanding:
            start = time.time() + 0.00001
            while time.time() - start < max_search_time:
                if status_update is not None:
                    if time.time() - status_start > status_update:
                        status = self.status()
                        logger.info(status, extra=self.logger_kwargs)
                        status_start = time.time()
                if self.expand_classes(1):
                    # this function returns True if no more classes to expand
                    expanding = False
                    logger.info("No more classes to expand.",
                                extra=self.logger_kwargs)
                    break
            start = time.time()
            if smallest:
                proof_tree = self.find_smallest_proof_tree()
            else:
                proof_tree = self.get_proof_tree()
            if proof_tree is not None:
                found_string = "Proof tree found {}\n".format(
                        time.strftime("%a, %d %b %Y %H:%M:%S", time.gmtime()))
                found_string += "Time taken was {} seconds\n\n".format(
                                                        self._time_taken)
                found_string += self.status()
                found_string += json.dumps(proof_tree.to_jsonable())
                logger.info(found_string, extra=self.logger_kwargs)
                if genf:
                    min_poly, func = proof_tree.get_min_poly(solve=True)
                    return proof_tree, min_poly, func
                return proof_tree
            # worst case, search every hour
            multiplier = 100 // perc
            max_search_time = min(multiplier*(time.time() - start), 3600)
            if max_time is not None:
                if self._time_taken > max_time:
                    logger.info(self.status(), extra=self.logger_kwargs)
                    if save:
                        string = "The universe: \n"
                        string += json.dumps(self.to_dict())
                        logger.info(string, extra=self.logger_kwargs)
                    logger.warn("Exceeded maximum time. Aborting auto search.",
                                extra=self.logger_kwargs)
                    return

    def has_proof_tree(self):
        """Return True if a proof tree has been found, false otherwise."""
        return self._has_proof_tree

    def tree_search_prep(self):
        """
        Return rule dictionary ready for tree searcher.

        Converts all rules to be in terms of equivalence database.
        """
        start_time = time.time()
        rules_dict = defaultdict(set)

        for rule in self.ruledb:
            self._add_rule_to_rules_dict(rule, rules_dict)

        for label in self.classdb.verified_labels():
            verified_label = self.equivdb[label]
            rules_dict[verified_label] |= set(((),))

        self.prep_for_tree_search_time += time.time() - start_time
        return rules_dict

    def _add_rule_to_rules_dict(self, rule, rules_dict):
        """Add a rule to given dictionary."""
        first, rest = rule
        eqv_first = self.equivdb[first]
        eqv_rest = tuple(sorted(self.equivdb[x] for x in rest))
        rules_dict[eqv_first] |= set((tuple(eqv_rest),))

    def equivalent_strategy_verified_label(self, label):
        """Return equivalent strategy verified label if one exists."""
        for eqv_label in self.equivdb.equivalent_set(label):
            if self.classdb.is_strategy_verified(eqv_label):
                return eqv_label

    def rule_from_equivence_rule(self, eqv_start, eqv_ends):
        """Return a rule that satisfies the equivalence rule."""
        for rule in self.ruledb:
            start, ends = rule
            if not self.equivdb.equivalent(start, eqv_start):
                continue
            if tuple(sorted(eqv_ends)) == tuple(sorted(self.equivdb[l]
                                                for l in ends)):
                return start, ends

    def find_tree(self):
        """Search for a random tree based on current data found."""
        start = time.time()

        rules_dict = self.tree_search_prep()
        # Prune all unverified labels (recursively)
        if self.iterative:
            rules_dict = iterative_prune(rules_dict,
                                         root=self.equivdb[self.start_label])
        else:
            rules_dict = prune(rules_dict)

        # only verified labels in rules_dict, in particular, there is a proof
        # tree if the start label is in the rules_dict
        for label in rules_dict.keys():
            self.equivdb.update_verified(label)

        if self.equivdb[self.start_label] in rules_dict:
            self._has_proof_tree = True
            if self.iterative:
                proof_tree = iterative_proof_tree_finder(
                                        rules_dict,
                                        root=self.equivdb[self.start_label])
            else:
                proof_tree = random_proof_tree(
                                        rules_dict,
                                        root=self.equivdb[self.start_label])
        else:
            proof_tree = None

        self.tree_search_time += time.time() - start
        self._time_taken += time.time() - start
        return proof_tree

    def get_proof_tree(self):
        """
        Return a random proof tree if one exists."""
        logger.debug("Searching for tree", extra=self.logger_kwargs)
        proof_tree_node = self.find_tree()
        if proof_tree_node is not None:
            proof_tree = ProofTree.from_comb_spec_searcher(proof_tree_node,
                                                           self)
            assert proof_tree is not None
            return proof_tree

    def all_proof_trees(self):
        """A generator that yields all proof trees in the universe."""
        start = time.time()
        root_label = self.equivdb[self.start_label]

        rules_dict = self.tree_search_prep()
        # Prune all unverified labels (recursively)
        if self.iterative:
            rules_dict = iterative_prune(rules_dict,
                                         root=root_label)
        else:
            rules_dict = prune(rules_dict)

        if self.equivdb[self.start_label] in rules_dict:
            if self.iterative:
                raise NotImplementedError("There is no method for yielding all"
                                          " iterative proof trees.")
            else:
                proof_trees = proof_tree_generator_dfs(
                                        rules_dict,
                                        root=root_label)
        else:
            logger.info("There are no proof trees.")
            return
        for proof_tree_node in proof_trees:
            yield ProofTree.from_comb_spec_searcher(proof_tree_node, self)

    def find_smallest_proof_tree(self):
        """Return a smallest proof tree in the universe. It uses exponential
        search to find it."""
        if self.iterative:
            raise NotImplementedError("There is no method for finding "
                                      " smallest iterative proof trees.")
        start = time.time()
        root_label = self.equivdb[self.start_label]
        logger.debug("Searching for tree", extra=self.logger_kwargs)
        rules_dict = self.tree_search_prep()
        rules_dict = prune(rules_dict)

        if self.equivdb[self.start_label] in rules_dict:
            bound = 1
            # Determine an upper bound on the size of a smallest proof tree.
            while True:
                logger.info("Looking for tree with max size {}".format(bound),
                            extra=self.logger_kwargs)
                try:
                    tree = next(proof_tree_generator_dfs(rules_dict,
                                                         root=root_label,
                                                         maximum=bound))
                    break
                except StopIteration:
                    bound *= 2
            minimum = 1
            maximum = bound
            # Binary search to find a smallest proof tree.
            while minimum < maximum:
                middle = (minimum + maximum) // 2
                logger.info("Looking for tree with max size {}".format(middle),
                            extra=self.logger_kwargs)
                try:
                    tree = next(proof_tree_generator_dfs(rules_dict,
                                                         root=root_label,
                                                         maximum=middle))
                    maximum = middle
                except StopIteration:
                    minimum = middle + 1
            return ProofTree.from_comb_spec_searcher(tree, self)
        else:
            logger.info("There are no proof trees.", extra=self.logger_kwargs)
            return



#!/usr/bin/env python
import os

from setuptools import find_packages, setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="comb_spec_searcher",
    version="0.2.2",
    author="Permuta Triangle",
    author_email="permutatriangle@gmail.com",
    description="A library for performing combinatorial exploration.",
    license="BSD-3",
    keywords="enumerative combinatorics combinatorial specification counting",
    url="https://github.com/PermutaTriangle/comb_spec_searcher",
    project_urls={
        'Source': 'https://github.com/PermutaTriangle/comb_spec_searcher',
        'Tracker': ('https://github.com/PermutaTriangle/comb_spec_searcher'
                    '/issues')
    },
    packages=find_packages(),
    long_description=read("README.rst"),
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: PyPy',

        'Topic :: Education',
        'Topic :: Scientific/Engineering :: Mathematics',
    ],
    install_requires=[
        'logzero==1.5.0',
        'sympy==1.4',
        'psutil==5.6.3'
    ],
    setup_requires=['pytest-runner==5.1'],
    tests_require=[
        'pytest==5.1.1',
        'pytest-cov==2.7.1',
        'pytest-isort==0.3.1',
        'pytest-pep8==1.0.6',
        'pytest-repeat==0.8.0',
        'docutils==0.15.2',
        'Pygments==2.4.2'
    ]
)



'''
Wrappers for combinatorial rules.

In the future, you should also declare the method needed for counting. The
types supported for a rule a -> b_1, b_2, ..., b_k are
    - disjoint union: f(a) = f(b_1) + f(b_2) + ... + f(b_k)
    - cartesiian product: f(a) = f(b_1) * f(b_2) * ... * f(b_k)
'''
import warnings
from collections import Iterable
from functools import partial


class Rule(object):
    def __init__(self, formal_step, comb_classes, inferable, possibly_empty,
                 workable, ignore_parent=False, constructor='disjoint'):
        if not isinstance(formal_step, str):
            raise TypeError("Formal step not a string")
        if not isinstance(comb_classes, Iterable):
            raise TypeError("The combinatorial classes are not iterable")
        if not comb_classes:
            raise TypeError(("There are no combinatorial classes, a strategy"
                             " contains a list of combinatorial classes"))
        if not isinstance(workable, Iterable):
            raise TypeError("Workable should an iterable")
        if not comb_classes:
            raise TypeError(("There are no combinatorial classes, a strategy"
                             " contains a list of combinatorial classes"))
        if any(not isinstance(x, bool) for x in workable):
            raise TypeError("Workable should be an iterable of booleans")
        if any(not isinstance(x, bool) for x in inferable):
            raise TypeError("Inferable should be an iterable of booleans")
        if constructor not in ['disjoint', 'cartesian', 'equiv', 'other']:
            raise ValueError(("Not valid constructor. Only accepts"
                              " disjoint or cartesian."))

        self.constructor = constructor
        self.formal_step = formal_step
        self.comb_classes = [comb_class for comb_class in comb_classes]
        self.inferable = [x for x in inferable]
        self.possibly_empty = [x for x in possibly_empty]
        self.workable = [x for x in workable]
        self.ignore_parent = ignore_parent


def BatchRule(formal_step, comb_classes):
    """A function for creating rules with the disjoint union constructor."""
    return Rule(formal_step, comb_classes,
                [True for _ in comb_classes],
                [True for _ in comb_classes],
                [True for _ in comb_classes],
                constructor='disjoint')


def DecompositionRule(formal_step, comb_classes, ignore_parent=True):
    """A function for creating rules with the cartesian product constructor."""
    return Rule(formal_step, comb_classes,
                [False for _ in comb_classes],
                [False for _ in comb_classes],
                [True for _ in comb_classes],
                ignore_parent=ignore_parent, constructor='cartesian')


def EquivalenceRule(formal_step, comb_class):
    """A function for creating equivalence rules."""
    return Rule(formal_step, [comb_class], [True], [False], [True],
                constructor='equiv')


def InferralRule(formal_step, comb_class):
    """A function for creating inferral rules."""
    return Rule(formal_step, [comb_class], [True], [False],
                [True], ignore_parent=True, constructor='equiv')


class VerificationRule(object):
    """A wrapper for verification rules."""
    def __init__(self, formal_step):
        """
        Constructor for verification rules.

        Formal step explains why verified.
        """
        if not isinstance(formal_step, str):
            raise TypeError("Formal step not a string")

        self.formal_step = formal_step


# Below is deprecated functions and classes for backwards compatability.


def BatchStrategy(formal_step, comb_classes):
    """A deprecated function for creating rules with the disjoint union
    constructor."""
    warnings.warn(("The BatchStrategy class is deprecated."
                   " Use BatchRule instead."),
                  DeprecationWarning, stacklevel=2)
    return BatchRule(formal_step, comb_classes)


def DecompositionStrategy(formal_step, comb_classes,
                          ignore_parent=True):
    """A deprecated function for creating rules with the cartesian product
    constructor."""
    warnings.warn(("The DecompositionStrategy function is deprecated."
                   " Use DecompositionRule instead."),
                  DeprecationWarning, stacklevel=2)
    return DecompositionRule(formal_step, comb_classes, ignore_parent)


def EquivalenceStrategy(formal_step, comb_class):
    """A deprecated function for creating equivalence rules."""
    warnings.warn(("The EquivalenceStrategy function is deprecated."
                   " Use EquivalenceRule instead."),
                  DeprecationWarning, stacklevel=2)
    return EquivalenceRule(formal_step, comb_class)


def InferralStrategy(formal_step, comb_class):
    """A deprecated function for creating inferral rules."""
    warnings.warn(("The InferralStrategy function is deprecated."
                   " Use InferralRule instead."),
                  DeprecationWarning, stacklevel=2)
    return InferralRule(formal_step, comb_class)


class Strategy(Rule):
    """A deprecated class for creating rules."""
    def __init__(self, *args, **kwargs):
        warnings.warn("The Strategy class is deprecated. Use Rule instead.",
                      DeprecationWarning, stacklevel=2)
        Rule.__init__(self, *args, **kwargs)


class VerificationStrategy(VerificationRule):
    """A deprecated wrapper for verification strategies."""

    def __init__(self, formal_step):
        """
        Constructor for verification strategies.

        Formal step explains why verified.
        """
        warnings.warn(("The VerificationStrategy class is deprecated."
                       " Use VerificationRule instead."),
                      DeprecationWarning, stacklevel=2)
        VerificationRule.__init__(self, formal_step)
